#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2022, 2024 Chris Lamb <lamby@debian.org>
# Copyright © 2021 Jean-Romain Garnier <salsa@jean-romain.com>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import pytest
import os.path
import subprocess

from diffoscope.config import Config
from diffoscope.comparators.ar import ArFile
from diffoscope.comparators.elf import ElfFile
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.directory import FilesystemDirectory
from diffoscope.comparators.missing_file import MissingFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import data, load_fixture, assert_diff
from ..utils.tools import (
    skip_unless_tools_exist,
    skip_if_binutils_does_not_support_x86,
    skip_unless_module_exists,
    skip_if_tool_version_is,
    skip_unless_tool_is_at_least,
)
from .test_rlib import llvm_version

obj1 = load_fixture("test1.o")
obj2 = load_fixture("test2.o")
ignore_readelf_errors1 = load_fixture("test1.debug")
ignore_readelf_errors2 = load_fixture("test2.debug")


@pytest.fixture(scope="function", autouse=True)
def init_tests(request, monkeypatch):
    # Ignore radare2 commands so decompiling is skipped
    # See test_elf_decompiler.py for tests related to decompiler
    monkeypatch.setattr(Config(), "exclude_commands", ["^radare2.*"])


def readelf_version():
    try:
        out = subprocess.check_output(["readelf", "--version"])
    except subprocess.CalledProcessError as e:
        out = e.output

    # Only match GNU readelf; we only need to match some versions
    m = re.match(r"^GNU readelf .* (?P<version>[\d.]+)\n", out.decode("utf-8"))

    # Return '0' as the version if we can't parse one; it should be harmless.
    if m is None:
        return "0"

    return m.group("version")


def test_obj_identification(obj1):
    assert isinstance(obj1, ElfFile)


def test_obj_no_differences(obj1):
    difference = obj1.compare(obj1)
    assert difference is None


@pytest.fixture
def obj_differences(obj1, obj2):
    return obj1.compare(obj2).details


@skip_unless_tools_exist("readelf")
@skip_if_tool_version_is("readelf", readelf_version, "2.29")
@skip_if_binutils_does_not_support_x86()
def test_obj_compare_non_existing(monkeypatch, obj1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = obj1.compare(MissingFile("/nonexisting", obj1))
    assert difference.source2 == "/nonexisting"
    assert len(difference.details) > 0


@skip_unless_tools_exist("readelf")
@skip_unless_tool_is_at_least("readelf", readelf_version, "2.38.50")
@skip_if_binutils_does_not_support_x86()
def test_diff(obj_differences):
    assert len(obj_differences) == 1
    assert_diff(obj_differences[0], "elf_obj_expected_diff")


TEST_LIB1_PATH = data("test1.a")
TEST_LIB2_PATH = data("test2.a")


@pytest.fixture
def lib1():
    return specialize(FilesystemFile(TEST_LIB1_PATH))


@pytest.fixture
def lib2():
    return specialize(FilesystemFile(TEST_LIB2_PATH))


def test_lib_identification(lib1):
    assert isinstance(lib1, ArFile)


def test_lib_no_differences(lib1):
    difference = lib1.compare(lib1)
    assert difference is None


@pytest.fixture
def lib_differences(lib1, lib2):
    return lib1.compare(lib2).details


@skip_unless_tools_exist("readelf", "objdump")
@skip_unless_tool_is_at_least("readelf", readelf_version, "2.38.50")
@skip_if_binutils_does_not_support_x86()
def test_lib_differences(lib_differences):
    assert len(lib_differences) == 2
    assert lib_differences[0].source1 == "file list"
    assert_diff(lib_differences[0], "elf_lib_metadata_expected_diff")
    assert "objdump" in lib_differences[1].details[0].source1
    assert_diff(lib_differences[1].details[0], "elf_lib_objdump_expected_diff")


@skip_unless_tools_exist("readelf", "objdump")
@skip_if_tool_version_is("readelf", readelf_version, "2.29")
@skip_if_binutils_does_not_support_x86()
def test_lib_compare_non_existing(monkeypatch, lib1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = lib1.compare(MissingFile("/nonexisting", lib1))
    assert difference.source2 == "/nonexisting"
    assert len(difference.details) > 0


TEST_LIBMIX1_PATH = data("elfmix1.not_a")
TEST_LIBMIX2_PATH = data("elfmix2.a")


@pytest.fixture
def libmix1():
    return specialize(FilesystemFile(TEST_LIBMIX1_PATH))


@pytest.fixture
def libmix2():
    return specialize(FilesystemFile(TEST_LIBMIX2_PATH))


@pytest.fixture
def libmix_differences(libmix1, libmix2):
    return libmix1.compare(libmix2).details


@skip_unless_tools_exist("xxd")
@skip_unless_tools_exist("llvm-readobj", "llvm-objdump")
@skip_unless_tools_exist("readelf", "objdump")
@skip_unless_tool_is_at_least("readelf", readelf_version, "2.38.50")
@skip_if_binutils_does_not_support_x86()
def test_libmix_differences(libmix_differences):
    assert len(libmix_differences) == 5
    file_list, mach_o, x86_o, src_c, x_obj = libmix_differences

    # Check order and basic identification
    assert file_list.source1 == "file list"
    x86_o = x86_o.details[0]
    assert x86_o.source1.startswith("objdump ")
    assert src_c.source1.endswith(".c")

    mach_o = mach_o.details[0]
    for diff in mach_o.details:
        assert diff.source1.startswith("llvm-objdump ")

    # Content
    assert "return42_or_3" in file_list.unified_diff
    assert_diff(x86_o, "elfmix_disassembly_expected_diff")
    assert_diff(src_c, "elfmix_src_c_expected_diff")

    if llvm_version() < "16":
        mach_o_filenames = ["elfmix_mach_o_expected_diff__text"]
    else:
        mach_o_filenames = ["elfmix_mach_o_expected_diff__text_16"]

    for idx, diff in enumerate(mach_o.details):
        assert_diff(diff, mach_o_filenames[idx])

    x_obj = x_obj.details[0]
    if x_obj.source1.startswith("readelf "):
        assert_diff(x_obj, "elfmix_x_obj_expected_diff")
    elif x_obj.source1.startswith("objdump "):
        assert_diff(x_obj, "elfmix_x_obj_objdump_expected_diff")
    else:
        pytest.fail(
            f"x_obj is neither readelf or objdump: {repr(x_obj.source1)}"
        )


TEST_DBGSYM_DEB1_PATH = data("dbgsym/add/test-dbgsym_1_amd64.deb")
TEST_DBGSYM_DEB2_PATH = data("dbgsym/mult/test-dbgsym_1_amd64.deb")


@pytest.fixture
def dbgsym_dir1():
    container = FilesystemDirectory(
        os.path.dirname(TEST_DBGSYM_DEB1_PATH)
    ).as_container
    return specialize(
        FilesystemFile(TEST_DBGSYM_DEB1_PATH, container=container)
    )


@pytest.fixture
def dbgsym_dir2():
    container = FilesystemDirectory(
        os.path.dirname(TEST_DBGSYM_DEB2_PATH)
    ).as_container
    return specialize(
        FilesystemFile(TEST_DBGSYM_DEB2_PATH, container=container)
    )


@pytest.fixture
def dbgsym_differences(monkeypatch, dbgsym_dir1, dbgsym_dir2):
    monkeypatch.setattr(Config(), "use_dbgsym", "yes")
    return dbgsym_dir1.compare(dbgsym_dir2)


@skip_unless_tools_exist("readelf", "objdump", "objcopy", "xz")
@skip_if_binutils_does_not_support_x86()
@skip_unless_module_exists("debian.deb822")
def test_differences_with_dbgsym(dbgsym_differences):
    assert dbgsym_differences.details[2].source1 == "data.tar.xz"
    bin_details = dbgsym_differences.details[2].details[0].details[0]
    assert bin_details.source1 == "./usr/bin/test"
    assert bin_details.details[1].source1.startswith("objdump")
    assert (
        "test-cases/dbgsym/package/test.c:2"
        in bin_details.details[1].unified_diff
    )


@skip_unless_tools_exist("readelf", "objdump", "objcopy", "xz")
@skip_if_binutils_does_not_support_x86()
@skip_unless_module_exists("debian.deb822")
def test_original_gnu_debuglink(dbgsym_differences):
    bin_details = dbgsym_differences.details[2].details[0].details[0]
    assert ".gnu_debuglink" in bin_details.details[2].source1
    assert_diff(bin_details.details[2], "gnu_debuglink_expected_diff")


def test_ignore_readelf_errors1_identify(ignore_readelf_errors1):
    assert isinstance(ignore_readelf_errors1, ElfFile)


def test_ignore_readelf_errors2_identify(ignore_readelf_errors2):
    assert isinstance(ignore_readelf_errors2, ElfFile)


@pytest.fixture
def ignore_readelf_errors_differences(
    ignore_readelf_errors1, ignore_readelf_errors2
):
    return ignore_readelf_errors1.compare(ignore_readelf_errors2).details


@skip_unless_tools_exist("readelf")
@skip_if_tool_version_is("readelf", readelf_version, "2.29")
@skip_if_binutils_does_not_support_x86()
def test_ignore_readelf_errors(ignore_readelf_errors_differences):
    assert_diff(
        ignore_readelf_errors_differences[0],
        "ignore_readelf_errors_expected_diff",
    )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017 Ximin Luo <infinity0@debian.org>
# Copyright © 2017-2021, 2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

from diffoscope.tools import tool_required
from diffoscope.tempfiles import get_temporary_directory
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command

import shutil
import os.path
import logging
import binascii

RDS_HEADERS = {
    binascii.a2b_hex("580a000000020003"),
    binascii.a2b_hex("580a000000030003"),
    binascii.a2b_hex("580a000000030004"),
}

DUMP_RDB = rb"""
hideOutput = lazyLoad(commandArgs(TRUE));

safeDeparse <- function(obj) {
    tryCatch({
        deparse(obj)
    }, error = function (e) {
        (deparse(e))
    });
}

for (x in ls(all.names = TRUE, sorted = TRUE)) {
    obj = get(x)

    cat(sprintf("%s (%s) = ", x, typeof(obj)), sep = "");

    if (typeof(obj) == "environment") {
        cat("\n{\n", sep = "");
        for (y in ls(obj, all.names = TRUE, sorted = TRUE)) {
            obj2 = get(y, envir = obj);
            cat(sprintf("    \"%s\" = \"%s\"\n", y, safeDeparse(obj2)), sep = "");
        }
        cat("}\n");
    } else {
        for (line in safeDeparse(obj))
            cat(line, "\n", sep = "");
    }
    cat("\n");
}
"""

logger = logging.getLogger(__name__)


def check_rds_extension(f):
    return f.name.endswith(".rds") or f.name.endswith(".rdx")


def get_module_path_for_rdb(rdb, temp_dir, module_name):
    """
    R's lazyLoad method does not take a filename directly to an .rdb file (eg.
    `/path/to/foo.rdb`) but rather the path without any extension (eg.
    `/path/to/foo`). It also requires that the .rdx file exists at
    `/path/to/foo.rdx`.

    We thus locate the corresponding .rdx file in the surrounding container and
    copy that to `foo.rdx`. We use a temporary directory to ensure we do not
    add files to the user's filesystem in the case of directly comparing two
    .rdb files or, worse, overwriting a file in its place.
    """

    # If we are not in a container, we will never be able to locate the
    # corresponding .rdx
    if rdb.container is None:
        return None

    # Calculate location of parallel .rdx file
    rdx_name = "{}.rdx".format(os.path.splitext(rdb.name)[0])

    # Query the container for the full path (eg. `./path/to/foo.rdx`)...
    try:
        rdx = rdb.container.get_member(rdx_name)
    except KeyError:
        # There is no guarantee that the equivalent .rdx file exists.
        # (see Debian:#1066991)
        return None

    if not os.path.exists(rdx.path):
        # ... falling back to looking in the same directory for when
        # comparing two .rdb files directly
        try:
            rdx = rdb.container.get_member(os.path.basename(rdx_name))
        except KeyError:
            return None

    if not os.path.exists(rdx.path):
        # Corresponding .rdx does not exist
        return None

    prefix = os.path.join(temp_dir, module_name)

    logger.debug("Copying %s and %s to %s", rdx.path, rdb.path, temp_dir)
    shutil.copy(rdb.path, f"{prefix}.rdb")
    shutil.copy(rdx.path, f"{prefix}.rdx")

    # Return the "module" path, ie. without an extension
    return prefix


class RdsReader(Command):
    @tool_required("Rscript")
    def cmdline(self):
        return [
            "Rscript",
            "--vanilla",
            "-e",
            "args <- commandArgs(TRUE); readRDS(args[1])",
            self.path,
        ]


class RdsFile(File):
    DESCRIPTION = "GNU R Rscript files (.rds)"

    @classmethod
    def recognizes(cls, file):
        if (
            check_rds_extension(file)
            or file.container
            and check_rds_extension(file.container.source)
        ):
            return file.file_header[:8] in RDS_HEADERS
        return False

    def compare_details(self, other, source=None):
        return [Difference.from_operation(RdsReader, self.path, other.path)]


class RdbReader(Command):
    MASK_STDERR = True

    @tool_required("Rscript")
    def cmdline(self):
        return ["Rscript", "--vanilla", "-", self.path]

    def input(self):
        return DUMP_RDB


class RdbFile(File):
    DESCRIPTION = "GNU R database files (.rdb)"
    FILE_EXTENSION_SUFFIX = {".rdb"}

    @classmethod
    def recognizes(cls, file):
        if not super().recognizes(file):
            return False

        # file(1) does not yet support identifying R database files, so we
        # match on "data" AND the filename ends with .rdb (via
        # FILE_EXTENSION_SUFFIX above) to avoid matching, say, Redis database
        # dumps (dump.rdb).
        return file.magic_file_type == "data"

    def compare_details(self, other, source=None):
        with get_temporary_directory(suffix="rdb") as tmpdir:
            a = get_module_path_for_rdb(self, tmpdir, "a")
            b = get_module_path_for_rdb(other, tmpdir, "b")

            if a is None or b is None:
                return []

            return [Difference.from_operation(RdbReader, a, b)]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2019-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.comparators.ffprobe import FfprobeFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least
from ..utils.nonexisting import assert_non_existing

mp3_1 = load_fixture("test1.mp3")
mp3_2 = load_fixture("test2.mp3")


def ffprobe_version():
    return (
        subprocess.check_output(["ffprobe", "-version"])
        .decode("utf-8")
        .split()[2]
    )


def test_identification(mp3_1):
    assert isinstance(mp3_1, FfprobeFile)


def test_no_differences(mp3_1):
    difference = mp3_1.compare(mp3_1)
    assert difference is None


@pytest.fixture
def differences(mp3_1, mp3_2):
    return mp3_1.compare(mp3_2).details


@skip_unless_tools_exist("ffprobe")
@skip_unless_tool_is_at_least("ffprobe", ffprobe_version, "4.4")
def test_diff(differences):
    # >= 7 adds a "(mp3float)" annotation
    computed_diff = differences[0].unified_diff.replace(
        "mp3 (mp3float)", "mp3"
    )
    assert computed_diff == get_data("mp3_expected_diff")


@skip_unless_tools_exist("ffprobe")
def test_compare_non_existing(monkeypatch, mp3_1):
    assert_non_existing(monkeypatch, mp3_1, has_null_source=False)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2020, 2022-2023 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import pytest

from diffoscope.main import main
from diffoscope.readers import load_diff_from_path
from diffoscope.presenters.utils import (
    create_limited_print_func,
    PrintLimitReached,
    PartialString,
)
from diffoscope.presenters.json import JSONPresenter

from .utils import diff_expand
from .utils.data import cwd_data, data, get_data
from .utils.tools import (
    skip_unless_tools_exist,
    skip_unless_file_version_is_at_least,
)

re_html = re.compile(
    r'.*<body(?P<body>.*)<div class="footer">', re.MULTILINE | re.DOTALL
)


def run(capsys, *args, pair=("test1.tar", "test2.tar")):
    with pytest.raises(SystemExit) as exc, cwd_data():
        main(args + pair)
    out, err = capsys.readouterr()

    assert err == ""
    assert exc.value.code == 1
    return out


def run_images(capsys, *args):
    return run(capsys, *args, pair=("test1.png", "test2.png"))


def extract_body(val):
    """
    Extract the salient parts of HTML fixtures that won't change between
    versions, etc.
    """

    result = re_html.search(val).group("body")

    # Ensure that we extracted something
    assert len(result) > 0

    return result


def expand_collapsed_json(tmpdir, name):
    diff = load_diff_from_path(data(name + ".collapsed-diff.json"))
    diff_path = str(tmpdir.join(name + ".diff.json"))
    with open(diff_path, "w") as fp:
        JSONPresenter(lambda x: print(x, file=fp)).start(
            diff.fmap(diff_expand)
        )
    return diff_path


def test_text_option_is_default(capsys):
    out = run(capsys)

    assert out == get_data("output.txt")


@skip_unless_file_version_is_at_least("5.44")
def test_text_proper_indentation(capsys):
    out = run(capsys, pair=("archive1.tar", "archive2.tar"))

    assert out == get_data("archive12.diff.txt")


def test_text_option_color(capsys):
    out = run(capsys, "--text-color=always")

    assert out == get_data("output.colored.txt")


def test_text_option_with_file(tmpdir, capsys):
    report_path = str(tmpdir.join("report.txt"))

    out = run(capsys, "--text", report_path)

    assert out == ""

    with open(report_path, "r", encoding="utf-8") as f:
        assert f.read() == get_data("output.txt")


def test_text_option_with_stdout(capsys):
    out = run(capsys, "--text", "-")

    assert out == get_data("output.txt")


def test_markdown(capsys):
    out = run(capsys, "--markdown", "-")

    assert out == get_data("output.md")


def test_restructuredtext(capsys):
    out = run(capsys, "--restructured-text", "-")

    assert out == get_data("output.rst")


def test_json(capsys):
    out = run(capsys, "--json", "-")

    assert out == get_data("output.json")


def test_no_report_option(capsys):
    out = run(capsys)

    assert out == get_data("output.txt")


def test_html_option_with_file(tmpdir, capsys):
    report_path = str(tmpdir.join("report.html"))

    out = run(capsys, "--html", report_path)

    assert out == ""
    with open(report_path, "r", encoding="utf-8") as f:
        body = extract_body(f.read())
        assert body.count('div class="difference"') == 4


@skip_unless_tools_exist("compare", "convert", "sng")
def test_html_visuals(tmpdir, capsys):
    report_path = str(tmpdir.join("report.html"))

    out = run_images(capsys, "--html", report_path)

    assert out == ""
    body = extract_body(open(report_path, "r", encoding="utf-8").read())
    assert '<img src="data:image/png;base64' in body
    assert '<img src="data:image/gif;base64' in body


def test_htmldir_option(tmpdir, capsys):
    html_dir = os.path.join(str(tmpdir), "target")

    out = run(capsys, "--html-dir", html_dir, "--jquery", "disable")

    assert out == ""
    assert os.path.isdir(html_dir)
    with open(
        os.path.join(html_dir, "index.html"), "r", encoding="utf-8"
    ) as f:
        body = extract_body(f.read())
        assert body.count('div class="difference"') == 4


def test_html_option_with_stdout(capsys):
    body = extract_body(run(capsys, "--html", "-"))

    assert body.count('div class="difference"') == 4


def test_html_regression_875281(tmpdir, capsys):
    diff_path = expand_collapsed_json(tmpdir, "debian-bug-875281")
    report_path = str(tmpdir.join("report.html"))
    out = run(
        capsys,
        "--html",
        report_path,
        "--max-page-size=5000",
        f"--load-existing-diff={diff_path}",
    )
    assert out == ""

    with open(report_path, "r", encoding="utf-8") as f:
        assert extract_body(f.read()) == get_data(
            "output_regression_875281.html"
        )


def test_html_presentation_issue_363(tmpdir, capsys):
    filename = "salsa-issue-363.diff.json"
    diff_data = get_data(filename)
    diff_path = str(tmpdir.join(filename))
    with open(diff_path, "w") as fp:
        fp.write(diff_data)
    report_path = str(tmpdir.join("report.html"))
    out = run(
        capsys,
        "--html",
        report_path,
        f"--load-existing-diff={diff_path}",
    )
    assert out == ""

    with open(report_path, "r", encoding="utf-8") as f:
        assert extract_body(f.read()) == get_data("output_expected_363.html")


def test_html_presentation_issue_372(tmpdir, capsys):
    filename = "salsa-issue-372.diff.json"
    diff_data = get_data(filename)
    diff_path = str(tmpdir.join(filename))
    with open(diff_path, "w") as fp:
        fp.write(diff_data)
    report_path = str(tmpdir.join("report.html"))
    out = run(
        capsys,
        "--html",
        report_path,
        f"--load-existing-diff={diff_path}",
    )
    assert out == ""

    with open(report_path, "r", encoding="utf-8") as f:
        assert extract_body(f.read()) == get_data("output_expected_372.html")


def test_limited_print():
    def fake(x):
        return None

    with pytest.raises(PrintLimitReached):
        p = create_limited_print_func(fake, 5)
        p("123456")
    with pytest.raises(PrintLimitReached):
        p = create_limited_print_func(fake, 5)
        p("123")
        p("456")
    p = create_limited_print_func(fake, 5)
    p("123")
    p("456", force=True)


def test_partial_string():
    a, b = object(), object()
    tmpl = PartialString("{0} {1}", a, b)
    assert tmpl.holes == (a, b)
    assert tmpl.format({a: "Hello,", b: "World!"}) == "Hello, World!"
    assert tmpl.pformat({a: "Hello,"}) == PartialString("Hello, {0}", b)
    assert tmpl.pformat({b: "World!"}) == PartialString("{0} World!", a)
    assert tmpl.base_len, tmpl.num_holes == (1, 2)
    assert tmpl.size(hole_size=33) == 67
    assert tmpl.pformat({a: PartialString("{0}", b)}) == PartialString(
        "{0} {0}", b
    )
    assert tmpl.pformat({a: tmpl}) == PartialString("{0} {1} {1}", a, b)
    assert tmpl.pformat({b: tmpl}) == PartialString("{0} {0} {1}", a, b)
    PartialString("{1}", a, b)
    with pytest.raises(IndexError):
        PartialString("{0} {1} {2}", a, b)


def test_partial_string_cont():
    t, cont = PartialString.cont()
    t = cont(t, "x: {0}\ny: {1}\n{-1}", object(), object())
    t = cont(t, "z: {0}\n{-1}", object())
    t = cont(t, "")
    key = t.holes
    assert (
        t.format({key[0]: "line1", key[1]: "line2", key[2]: "line3"})
        == "x: line1\ny: line2\nz: line3\n"
    )
    assert t.size(hole_size=5) == 27


def test_partial_string_numl():
    tmpl = PartialString.numl("{0} {1} {2}", 2, object())
    assert tmpl.holes[:2] == (0, 1)
    assert tmpl.pformatl("(1)", "(2)", "(o)") == PartialString("(1) (2) (o)")


def test_partial_string_escape():
    tmpl = PartialString.numl(
        "find {0} -name {1} " + PartialString.escape(r"-exec ls -la {} \;"), 2
    )
    assert tmpl == PartialString(
        "find {0} -name {1} -exec ls -la {{}} \\;", *tmpl.holes
    )
    assert tmpl.size() == 33
    assert tmpl.size(4) == 39
    assert tmpl == PartialString.numl(
        r"find {0} -name {1} -exec ls -la {2} \;", 3
    ).pformat({2: "{}"})

    assert tmpl.pformatl("my{}path", "my{}file") == PartialString(
        "find my{{}}path -name my{{}}file -exec ls -la {{}} \\;"
    )
    assert (
        tmpl.formatl("my{}path", "my{}file")
        == "find my{}path -name my{}file -exec ls -la {} \\;"
    )

    esc = PartialString("{{}} {0}", None)
    assert esc.pformat({None: PartialString.of(None)}) == esc
    assert esc.format({None: "0"}) == "{} 0"
    with pytest.raises(ValueError):
        PartialString("{}")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Reiner Herrmann <reiner@reiner-h.de>
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import os
import tempfile

from diffoscope.config import Config
from diffoscope.comparators.missing_file import MissingFile
from diffoscope.comparators.fsimage import FsImageFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist, skip_unless_module_exists

img1 = load_fixture("test1.ext4")
img2 = load_fixture("test2.ext4")
img1_fat12 = load_fixture("test1.fat12")
img2_fat12 = load_fixture("test2.fat12")
img1_fat16 = load_fixture("test1.fat16")
img1_fat32 = load_fixture("test1.fat32")


@pytest.fixture(scope="session")
def guestfs_tempdir():
    import guestfs

    with tempfile.TemporaryDirectory(suffix="_diffoscope") as cachedir:
        g = guestfs.GuestFS(python_return_dict=True)
        g.set_cachedir(cachedir)
        # set cachedir for the diffoscope.comparators.fsimage module as well
        os.environ["LIBGUESTFS_CACHEDIR"] = cachedir
        g.add_drive_opts("/dev/null", format="raw", readonly=1)
        try:
            g.launch()
        except RuntimeError as e:
            # to debug this, set LIBGUESTFS_DEBUG=1 and give -s to pytest.
            # Unfortunately we can't capture the logs, due to capsys not being
            # available in a module/session scope, and g.set_event_callback
            # segfaults on my system.
            pytest.skip("guestfs not working on the system: %r" % e)
        yield cachedir


def test_identification(img1):
    assert isinstance(img1, FsImageFile)


def test_identification_fat12(img1_fat12):
    assert isinstance(img1_fat12, FsImageFile)


def test_identification_fat16(img1_fat16):
    assert isinstance(img1_fat16, FsImageFile)


def test_identification_fat32(img1_fat32):
    assert isinstance(img1_fat32, FsImageFile)


@skip_unless_tools_exist("qemu-img")
@skip_unless_module_exists("guestfs")
def test_no_differences(img1, guestfs_tempdir):
    difference = img1.compare(img1)
    assert difference is None


@pytest.fixture
def differences(img1, img2, guestfs_tempdir):
    return img1.compare(img2).details


@skip_unless_tools_exist("qemu-img")
@skip_unless_module_exists("guestfs")
def test_differences(differences, guestfs_tempdir):
    assert differences[0].source1 == "test1.ext4.tar"
    tarinfo = differences[0].details[0]
    tardiff = differences[0].details[1]
    encodingdiff = tardiff.details[0]
    assert tarinfo.source1 == "file list"
    assert tarinfo.source2 == "file list"
    assert tardiff.source1 == "./date.txt"
    assert tardiff.source2 == "./date.txt"
    assert encodingdiff.source1 == "encoding"
    assert encodingdiff.source2 == "encoding"
    expected_diff = get_data("ext4_expected_diffs")
    found_diff = (
        tarinfo.unified_diff + tardiff.unified_diff + encodingdiff.unified_diff
    )
    assert expected_diff == found_diff


@skip_unless_tools_exist("qemu-img")
@skip_unless_module_exists("guestfs")
def test_compare_non_existing(monkeypatch, img1, guestfs_tempdir):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = img1.compare(MissingFile("/nonexisting", img1))
    assert difference.source2 == "/nonexisting"
    assert difference.details[-1].source2 == "/dev/null"


@pytest.fixture
def differences_fat(img1_fat12, img2_fat12, guestfs_tempdir):
    return img1_fat12.compare(img2_fat12).details


@skip_unless_tools_exist("qemu-img")
@skip_unless_module_exists("guestfs")
def test_differences_fat(differences_fat, guestfs_tempdir):
    assert differences_fat[0].source1 == "filetype from file(1)"
    assert differences_fat[1].source1 == "test1.fat12.tar"
    tarinfo = differences_fat[1].details[0]
    tardiff = differences_fat[1].details[1]
    assert tarinfo.source1 == "file list"
    assert tarinfo.source2 == "file list"
    assert tardiff.source1 == "./test1.txt"
    assert tardiff.source2 == "./test1.txt"
    expected_diff = get_data("fat12_expected_diffs")
    found_diff = (
        differences_fat[0].unified_diff
        + tarinfo.unified_diff
        + tardiff.unified_diff
    )
    # workaround for file(1) bug in stretch
    found_diff = found_diff.replace("32 MB) ,", "32 MB),")
    assert expected_diff == found_diff



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Reiner Herrmann <reiner@reiner-h.de>
# Copyright © 2016-2017, 2019-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.


import logging
import time

logger = logging.getLogger(__name__)


class defaultint(int):
    pass


# Avoid setting values on this anywhere other than main.run_diffoscope(),
# otherwise tests may fail unpredictably depending on order-of-execution.
class Config:
    _singleton = {}

    def __init__(self):
        self.__dict__ = self._singleton
        if not self._singleton:
            self.reset()

    def reset(self):
        self.diff_context = 7

        self.max_diff_input_lines = 2**22
        self.max_diff_block_lines_saved = float("inf")

        # hard limits, restricts single-file and multi-file formats
        self.max_report_size = defaultint(40 * 2**20)  # 40 MB
        self.max_diff_block_lines = defaultint(2**10)  # 1024 lines
        # structural limits, restricts single-file formats
        # semi-restricts multi-file formats
        self.max_page_size = defaultint(40 * 2**20)  # 4 MB
        self.max_page_diff_block_lines = defaultint(2**7)  # 128 lines

        self.max_text_report_size = 0

        self.difftool = None
        self.diff_masks = ()
        self.new_file = False
        self.fuzzy_threshold = 110
        self.enforce_constraints = True
        self.excludes = ()
        self.exclude_commands = ()
        self.exclude_directory_metadata = "no"
        self.extended_filesystem_attributes = False
        self.compute_visual_diffs = False
        self.max_container_depth = 50
        self.use_dbgsym = "auto"
        self.force_details = False

        self.started = time.time()
        self.timeout = float("inf")

    def __setattr__(self, k, v):
        super(Config, self).__setattr__(k, v)

    def check_ge(self, a, b):
        va = getattr(self, a)
        vb = getattr(self, b)
        if va < vb:
            if not isinstance(vb, defaultint):
                raise ValueError(
                    "{0} ({1}) cannot be smaller than {2} ({3})".format(
                        a, va, b, vb
                    )
                )

            logger.warning(
                "%s (%s) < default value of %s (%s), setting latter to %s",
                a,
                va,
                b,
                vb,
                va,
            )
            setattr(self, b, va)

    def check_constraints(self):
        self.check_ge("max_diff_block_lines", "max_page_diff_block_lines")
        self.check_ge("max_report_size", "max_page_size")

    def timeout_exceeded(self):
        return (self.started + self.timeout) < time.time()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.sqlite import Sqlite3Database

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing

sqlite3db1 = load_fixture("test1.sqlite3")
sqlite3db2 = load_fixture("test2.sqlite3")


def test_identification(sqlite3db1):
    assert isinstance(sqlite3db1, Sqlite3Database)


def test_no_differences(sqlite3db1):
    difference = sqlite3db1.compare(sqlite3db1)
    assert difference is None


@pytest.fixture
def differences(sqlite3db1, sqlite3db2):
    return sqlite3db1.compare(sqlite3db2).details


@skip_unless_tools_exist("sqlite3")
def test_diff(differences):
    expected_diff = get_data("sqlite3_expected_diff")
    actual_diff = differences[0].unified_diff
    assert (
        actual_diff == expected_diff
        or actual_diff == expected_diff.replace('"test"', "test")
    )


@skip_unless_tools_exist("sqlite3")
def test_compare_non_existing(monkeypatch, sqlite3db1):
    assert_non_existing(monkeypatch, sqlite3db1, has_null_source=False)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2020 Chris Lamb <lamby@debian.org>
# Copyright © 2021 Brent Spillner <s p i l l n e r @ a c m . o r g>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import socket
import pytest

from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.socket_or_fifo import SocketOrFIFO, format_socket
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import get_data, load_fixture

sample_textfile = "text_ascii1"
sampletext = load_fixture(sample_textfile)


def make_socket(path):
    if os.path.exists(path):
        os.remove(path)
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.bind(path)
    return specialize(FilesystemFile(path))


def make_pipe(path):
    if os.path.exists(path):
        os.remove(path)
    os.mkfifo(path)
    return specialize(FilesystemFile(path))


@pytest.fixture
def endpoints(tmpdir):
    def makename(tag):
        return os.path.join(tmpdir, "test_" + tag)

    test_points = zip(
        [make_socket, make_socket, make_pipe, make_pipe],
        map(makename, ["socket1", "socket2", "pipe1", "pipe2"]),
    )
    yield [(name, f(name)) for (f, name) in test_points]
    for _, name in test_points:
        os.remove(name)


@pytest.fixture
def expected_results(endpoints):
    descriptions = [
        format_socket(obj.get_type(), path) for (path, obj) in endpoints
    ]
    [sock1_desc, sock2_desc, pipe1_desc, pipe2_desc] = descriptions

    # Prefix every line of the sample text file with '+' to predict RHS of the diff
    sampletext_contents = get_data(sample_textfile)
    sample_lines = sampletext_contents.count("\n")
    added_text = "+" + "\n+".join(sampletext_contents.split("\n")[:-1]) + "\n"

    sock_text_diff = (
        "@@ -1 +1,{} @@\n".format(sample_lines) + "-" + sock1_desc + added_text
    )
    pipe_text_diff = (
        "@@ -1 +1,{} @@\n".format(sample_lines) + "-" + pipe1_desc + added_text
    )
    sock_sock_diff = "@@ -1 +1 @@\n" + "-" + sock1_desc + "+" + sock2_desc
    pipe_pipe_diff = "@@ -1 +1 @@\n" + "-" + pipe1_desc + "+" + pipe2_desc
    sock_pipe_diff = "@@ -1 +1 @@\n" + "-" + sock1_desc + "+" + pipe1_desc
    pipe_sock_diff = "@@ -1 +1 @@\n" + "-" + pipe1_desc + "+" + sock1_desc
    yield (
        sock_text_diff,
        pipe_text_diff,
        sock_sock_diff,
        pipe_pipe_diff,
        sock_pipe_diff,
        pipe_sock_diff,
    )


def test_sockets(endpoints, expected_results, sampletext):
    names, objects = zip(*endpoints)
    sock1, sock2, pipe1, pipe2 = objects
    (
        sock_text_diff,
        pipe_text_diff,
        sock_sock_diff,
        pipe_pipe_diff,
        sock_pipe_diff,
        pipe_sock_diff,
    ) = expected_results

    assert isinstance(sock1, SocketOrFIFO)
    assert isinstance(pipe1, SocketOrFIFO)

    assert sock1.compare(sampletext).unified_diff == sock_text_diff
    assert pipe1.compare(sampletext).unified_diff == pipe_text_diff

    assert sock1.compare(sock1) == None
    assert pipe1.compare(pipe1) == None
    assert sock1.compare(sock2).unified_diff == sock_sock_diff
    assert pipe1.compare(pipe2).unified_diff == pipe_pipe_diff
    assert sock1.compare(pipe1).unified_diff == sock_pipe_diff
    assert pipe1.compare(sock1).unified_diff == pipe_sock_diff



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2022 Christopher Baines <mail@cbaines.net>
# Copyright © 2022-2023 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import shutil
import pytest

from diffoscope.comparators.lzip import LzipFile
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing

lzip1 = load_fixture("test1.lzip")
lzip2 = load_fixture("test2.lzip")


def test_identification(lzip1):
    assert isinstance(lzip1, LzipFile)


def test_no_differences(lzip1):
    difference = lzip1.compare(lzip1)
    assert difference is None


@pytest.fixture
def differences(lzip1, lzip2):
    return lzip1.compare(lzip2).details


@skip_unless_tools_exist("lzip")
def test_content_source(differences):
    assert differences[0].source1 == "test1"
    assert differences[0].source2 == "test2"


@skip_unless_tools_exist("lzip")
def test_content_source_without_extension(tmpdir, lzip1, lzip2):
    path1 = str(tmpdir.join("test1"))
    path2 = str(tmpdir.join("test2"))
    shutil.copy(lzip1.path, path1)
    shutil.copy(lzip2.path, path2)
    lzip1 = specialize(FilesystemFile(path1))
    lzip2 = specialize(FilesystemFile(path2))
    difference = lzip1.compare(lzip2).details
    assert difference[0].source1 == "test1-content"
    assert difference[0].source2 == "test2-content"


@skip_unless_tools_exist("lzip")
def test_content_diff(differences):
    assert_diff(differences[0], "text_ascii_expected_diff")


@skip_unless_tools_exist("lzip")
def test_compare_non_existing(monkeypatch, lzip1):
    assert_non_existing(monkeypatch, lzip1)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020, 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command


class H5dump(Command):
    @tool_required("h5dump")
    def cmdline(self):
        return ["h5dump", self.path]

    def filter(self, line):
        # Strip absolute `HDF5 "/path/to/filename.db"`
        return re.sub(
            r"^(HDF5) \"[^\"]+\"", "\\1", line.decode("utf-8")
        ).encode("utf-8")


class Hdf5File(File):
    DESCRIPTION = "Hierarchical Data Format database"
    FILE_TYPE_RE = re.compile(r"^Hierarchical Data Format \(version 5\) data")

    def compare_details(self, other, source=None):
        return [Difference.from_operation(H5dump, self.path, other.path)]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2020, 2023-2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.config import Config
from diffoscope.comparators.image import JPEGImageFile
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_between

image1 = load_fixture("test1.jpg")
image2 = load_fixture("test2.jpg")
image1_meta = load_fixture("test1_meta.jpg")
image2_meta = load_fixture("test2_meta.jpg")


def imagemagick_version(command):
    out = subprocess.check_output([command, "-version"]).decode("utf-8")
    # First line is expected to look like
    # "Version: ImageMagick 6.9.6-6 Q16 x86_64 20161125 ..."
    if not out.startswith("Version: ImageMagick "):
        return "0.0.0"
    return out.splitlines()[0].split()[2].strip()


def test_identification(image1):
    assert isinstance(image1, JPEGImageFile)


def test_no_differences(image1):
    difference = image1.compare(image1)
    assert difference is None


@pytest.fixture
def differences(image1, image2):
    return image1.compare(image2).details


@skip_unless_tools_exist("img2txt", "identify")
def test_diff(differences):
    assert_diff(differences[0], "jpeg_image_expected_diff")


@skip_unless_tools_exist("img2txt", "identify")
def test_compare_non_existing(monkeypatch, image1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = image1.compare(MissingFile("/nonexisting", image1))
    assert difference.source2 == "/nonexisting"
    assert len(difference.details) > 0


@pytest.fixture
def differences_meta(image1_meta, image2_meta):
    return image1_meta.compare(image2_meta).details


@skip_unless_tools_exist("img2txt", "identify")
@skip_unless_tool_is_between(
    "identify", lambda: imagemagick_version("identify"), "6.9.6", "7.0.0"
)
def test_diff_meta(differences_meta):
    assert_diff(differences_meta[-1], "jpeg_image_meta_expected_diff")


@skip_unless_tool_is_between(
    "convert", lambda: imagemagick_version("convert"), "6.9.6", "7.0.0"
)
@skip_unless_tools_exist("img2txt", "convert", "identify")
def test_has_visuals(monkeypatch, image1, image2):
    monkeypatch.setattr(Config(), "compute_visual_diffs", True)
    jpg_diff = image1.compare(image2)
    assert len(jpg_diff.details) == 2
    assert len(jpg_diff.details[0].visuals) == 2
    assert jpg_diff.details[0].visuals[0].data_type == "image/png;base64"
    assert jpg_diff.details[0].visuals[1].data_type == "image/gif;base64"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2017, 2019-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.comparators.squashfs import SquashfsFile

from ..utils.data import load_fixture, assert_diff, get_data
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least
from ..utils.nonexisting import assert_non_existing


def unsquashfs_version():
    # first line of 'unsquashfs -version' looks like:
    #   unsquashfs version 4.4pre-git (2019/08/15)
    try:
        out = subprocess.check_output(["unsquashfs", "-version"])
    except subprocess.CalledProcessError as e:
        out = e.output
    return out.decode("UTF-8").splitlines()[0].split()[2].strip()


squashfs1 = load_fixture("test1.squashfs")
squashfs2 = load_fixture("test2.squashfs")


def test_identification(squashfs1):
    assert isinstance(squashfs1, SquashfsFile)


def test_no_differences(squashfs1):
    difference = squashfs1.compare(squashfs1)
    assert difference is None


def test_no_warnings(capfd, squashfs1, squashfs2):
    _ = squashfs1.compare(squashfs2)
    _, err = capfd.readouterr()
    assert err == ""


@pytest.fixture
def differences(squashfs1, squashfs2):
    return squashfs1.compare(squashfs2).details


@skip_unless_tool_is_at_least("unsquashfs", unsquashfs_version, "4.4")
def test_superblock(differences):
    assert_diff(differences[0], "squashfs_superblock_expected_diff")


@skip_unless_tools_exist("unsquashfs")
def test_symlink(differences):
    assert differences[2].comment == "symlink"
    assert_diff(differences[2], "symlink_expected_diff")


@skip_unless_tools_exist("unsquashfs")
def test_compressed_files(differences):
    assert differences[3].source1 == "/text"
    assert differences[3].source2 == "/text"
    assert_diff(differences[3], "text_ascii_expected_diff")


@skip_unless_tools_exist("unsquashfs")
def test_compare_non_existing(monkeypatch, squashfs1):
    assert_non_existing(monkeypatch, squashfs1)


# Test things that require root

squashfs1_root = load_fixture("test1_root.squashfs")
squashfs2_root = load_fixture("test2_root.squashfs")


@pytest.fixture
def differences_root(squashfs1_root, squashfs2_root):
    return squashfs1_root.compare(squashfs2_root).details


@skip_unless_tools_exist("unsquashfs")
def test_symlink_root(differences_root):
    assert (
        get_data("squashfs_root_expected_diff")
        in differences_root[1].unified_diff
    )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2026 Chris Lamb <lamby@debian.org>
# Copyright © 2021 Jean-Romain Garnier <salsa@jean-romain.com>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

"""
The keys of this dictionary are filenames of executables (eg. `abootimg`)
that might resolve to, for example, `/usr/bin/abootimg`.
"""

EXTERNAL_TOOLS = {
    "7z": {"debian": "7zip", "guix": "7zip"},
    "aapt2": {"debian": "aapt"},
    "abootimg": {"debian": "abootimg", "guix": "abootimg"},
    "androguard": {"debian": "androguard"},
    "apksigcopier": {"debian": "apksigcopier"},
    "apktool": {"debian": "apktool"},
    "apksigner": {"debian": "apksigner"},
    "asar": {"arch": "asar"},
    "db_dump": {"debian": "db-util", "guix": "bdb"},
    "bsdtar": {
        "debian": "libarchive-tools",
        "arch": "libarchive",
        "fedora": "bsdtar",
        "FreeBSD": "libarchive",
        "guix": "libarchive",
    },
    "dumppdf": {"debian": "python3-pdfminer", "guix": "python-pdfminer-six"},
    "bzip2": {"debian": "bzip2", "arch": "bzip2", "guix": "bzip2"},
    "cbfstool": {"debian": "coreboot-utils"},
    "cd-iccdump": {
        "debian": "colord",
        "arch": "colord",
        "FreeBSD": "colord",
        "guix": "colord",
    },
    "cmp": {"debian": "diffutils", "arch": "diffutils", "guix": "diffutils"},
    "compare": {
        "debian": "imagemagick",
        "arch": "imagemagick",
        "guix": "imagemagick",
    },
    "cpio": {"debian": "cpio", "arch": "cpio", "guix": "cpio"},
    "dexdump": {"debian": "dexdump"},
    "diff": {"debian": "diffutils", "arch": "diffutils", "guix": "diffutils"},
    "docx2txt": {"debian": "docx2txt", "arch": "docx2txt", "guix": "docx2txt"},
    "dumpimage": {
        "debian": "u-boot-tools",
        "arch": "uboot-tools",
        "guix": "u-boot-tools",
    },
    "enjarify": {"debian": "enjarify", "arch": "enjarify", "guix": "enjarify"},
    "fdtdump": {
        "debian": "device-tree-compiler",
        "arch": "dtc",
        "guix": "dtc",
    },
    "ffprobe": {"debian": "ffmpeg", "arch": "ffmpeg", "guix": "ffmpeg"},
    "file": {"debian": "file", "arch": "file", "guix": "file"},
    "find": {"debian": "findutils", "arch": "findutils", "guix": "findutils"},
    "getfacl": {"debian": "acl", "arch": "acl", "guix": "acl"},
    "gifbuild": {
        "debian": "giflib-tools",
        "arch": "giflib",
        "guix": "giflib:bin",
    },
    "ghc": {"debian": "ghc", "arch": "ghc", "FreeBSD": "ghc", "guix": "ghc"},
    "gpg": {
        "debian": "gpg",
        "arch": "gnupg",
        "FreeBSD": "gnupg",
        "guix": "gnupg",
    },
    "guestfs": {"debian": "python3-guestfs"},
    "gzip": {"debian": "gzip", "arch": "gzip", "guix": "gzip"},
    "h5dump": {"debian": "hdf5-tools", "arch": "hdf5", "guix": "hdf5"},
    "html2text": {"debian": "html2text", "guix": "html2text"},
    "identify": {
        "debian": "imagemagick",
        "arch": "imagemagick",
        "guix": "imagemagick",
    },
    "img2txt": {
        "debian": "caca-utils",
        "arch": "libcaca",
        "FreeBSD": "libcaca",
        "guix": "libcaca",
    },
    "isoinfo": {
        "debian": "genisoimage",
        "arch": "cdrtools",
        "FreeBSD": "cdrtools",
        "guix": "cdrkit-libre",
    },
    "javap": {
        "debian": "default-jdk-headless | default-jdk | java-sdk",
        "arch": "java-environment",
        "guix": "openjdk:jdk",
    },
    "js-beautify": {
        "debian": "jsbeautifier",
        "arch": "python-jsbeautifier",
        "guix": "python-jsbeautifier",
    },
    "kbxutil": {"debian": "gnupg-utils", "guix": "gnupg"},
    "lipo": {},
    "llvm-bcanalyzer": {"debian": "llvm", "arch": "llvm", "guix": "llvm"},
    "llvm-config": {"debian": "llvm", "arch": "llvm"},
    "llvm-dis": {"debian": "llvm", "arch": "llvm", "guix": "llvm"},
    "llvm-readobj": {"debian": "llvm", "arch": "llvm", "guix": "llvm"},
    "llvm-objdump": {"debian": "llvm", "arch": "llvm", "guix": "llvm"},
    "ls": {"debian": "coreutils", "arch": "coreutils", "guix": "coreutils"},
    "lsattr": {
        "debian": "e2fsprogs",
        "arch": "e2fsprogs",
        "FreeBSD": "e2fsprogs",
        "guix": "e2fsprogs",
    },
    "lz4": {"debian": "lz4", "FreeBSD": "lz4", "guix": "lz4"},
    "lzip": {"debian": "lzip", "guix": "lzip"},
    "lzma": {"debian": "xz-utils", "arch": "xz", "guix": "xz"},
    "msgunfmt": {
        "debian": "gettext",
        "arch": "gettext",
        "FreeBSD": "gettext-tools",
        "guix": "gettext",
    },
    "convert": {
        "debian": "imagemagick",
        "arch": "imagemagick",
        "guix": "imagemagick",
    },
    "nm": {
        "debian": "binutils-multiarch",
        "arch": "binutils",
        "guix": "binutils",
    },
    "objcopy": {
        "debian": "binutils-multiarch",
        "arch": "binutils",
        "guix": "binutils",
    },
    "objdump": {
        "debian": "binutils-multiarch",
        "arch": "binutils",
        "guix": "binutils",
    },
    "ocamlobjinfo": {"debian": "ocaml", "arch": "ocaml", "guix": "ocaml"},
    "odt2txt": {"debian": "odt2txt", "arch": "odt2txt", "guix": "odt2txt"},
    "oggDump": {"debian": "oggvideotools"},
    "openssl": {"debian": "openssl", "arch": "openssl", "guix": "openssl"},
    "otool": {},
    "pgpdump": {"debian": "pgpdump", "arch": "pgpdump", "guix": "pgpdump"},
    "pdftotext": {
        "debian": "poppler-utils",
        "arch": "poppler",
        "FreeBSD": "poppler-utils",
        "guix": "poppler",
    },
    "pedump": {
        "debian": "mono-devel (>= 6.14.1+ds-3) | mono-utils (<< 6.14.1+ds-3)",
        "arch": "mono",
        "FreeBSD": "mono",
        "guix": "mono",
    },
    "ppudump": {
        "debian": "fp-utils",
        "arch": "fpc",
        "FreeBSD": "fpc",
        "guix": "fpc",
    },
    "ps2ascii": {
        "debian": "ghostscript",
        "arch": "ghostscript",
        "FreeBSD": "ghostscript",
        "guix": "ghostscript",
    },
    "pypdf": {"debian": "python3-pypdf"},
    "radare2": {"debian": "radare2", "arch": "radare2", "guix": "radare2"},
    "readelf": {
        "debian": "binutils-multiarch",
        "arch": "binutils",
        "guix": "binutils",
    },
    "rpm2cpio": {
        "debian": "rpm2cpio",
        "arch": "rpmextract",
        "FreeBSD": "rpm2cpio",
        "guix": "rpm",
    },
    "Rscript": {"debian": "r-base-core", "arch": "r", "guix": "r-minimal"},
    "showttf": {"debian": "fontforge-extras"},
    "sng": {"debian": "sng", "guix": "sng"},
    "ssconvert": {
        "debian": "gnumeric",
        "arch": "gnumeric",
        "guix": "gnumeric",
    },
    "ssh-keygen": {
        "debian": "openssh-client",
        "arch": "openssh",
        "guix": "openssh",
    },
    "stat": {"debian": "coreutils", "arch": "coreutils", "guix": "coreutils"},
    "strings": {"debian": "binutils-multiarch"},
    "sqlite3": {
        "debian": "sqlite3",
        "arch": "sqlite",
        "FreeBSD": "sqlite3",
        "guix": "sqlite",
    },
    "wasm2wat": {"debian": "wabt", "arch": "wabt", "guix": "wabt"},
    "tar": {"debian": "tar", "arch": "tar", "guix": "tar"},
    "tcpdump": {"debian": "tcpdump", "arch": "tcpdump", "guix": "tcpdump"},
    "ttx": {"debian": "fonttools", "guix": "python-fonttools"},
    "ukify": {
        "debian": "systemd-ukify",
        "arch": "systemd-ukify",
    },
    "unsquashfs": {
        "debian": "squashfs-tools",
        "arch": "squashfs-tools",
        "FreeBSD": "squashfs-tools",
        "guix": "squashfs-tools",
    },
    "xxd": {
        "debian": "xxd",
        "arch": "vim",
        "FreeBSD": "vim",
        "guix": "xxd",
    },
    "xz": {"debian": "xz-utils", "arch": "xz", "guix": "xz"},
    "zipinfo": {
        "debian": "unzip",
        "arch": "unzip",
        "FreeBSD": "unzip",
        "guix": "unzip",
    },
    "zipnote": {"debian": "zip", "guix": "zip"},
    "procyon": {"debian": "procyon-decompiler"},
    "dumpxsb": {"debian": "xmlbeans"},
    "xb-tool": {
        "debian": "libxmlb-utils",
        "arch": "libxmlb",
        "guix": "libxmlb",
    },
    "zipdetails": {"debian": "perl", "guix": "perl"},
    "zstd": {"debian": "zstd", "guix": "zstd"},
}

# May be populated at runtime by remapped names like
# readelf -> arm-none-eabi-readelf, etc
# diff -> gdiff, etc
REMAPPED_TOOL_NAMES = {}

# GNU programs whose names differ on some non-GNU systems such as FreeBSD etc
# AND where the CLI or output of the programs differ from the non-GNU system
# versions. In these cases, add them here and make sure you wrap uses of them
# in get_tool_name() to pick up the alternate names.
#
# If we only use POSIX CLI options and the output is identical to the system
# version (so that our tests don't break) then it's unnecessary to add it here.
GNU_TOOL_NAMES = {"diff", "readelf", "objcopy", "objdump"}

# Set of tools considered "large" in their installation size or too niche in
# their target users. This is so they can easily be excluded from installation
# if not required in most cases. Note that these are the names of the tools
# (ie. the keys in the EXTERNAL_TOOLS dict), not the package names.
HUGE_TOOLS = {
    "ghc",
    "ocamlobjinfo",
    "llvm-bcanalyzer",
    "llvm-config",
    "llvm-dis",
    "llvm-readobj",
    "llvm-objdump",
    "ppudump",
    "javap",
    "ssconvert",
    "apktool",
    "apksigner",
    "pedump",
    "radare2",
    "dumpxsb",
}



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2022, 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import subprocess

from diffoscope.difference import Difference
from diffoscope.exc import RequiredToolNotFound
from diffoscope.tools import tool_required

from .text import TextFile
from .utils.command import Command


class Htmltotext(Command):
    @tool_required("html2text")
    def cmdline(self):
        return ["html2text", self.path]


class HtmlFile(TextFile):
    DESCRIPTION = "HTML files (.html)"
    FILE_TYPE_RE = re.compile(r"^HTML document")

    def compare(self, other, source=None):
        difference = super().compare(other, source)

        if difference is None:
            return difference

        # Show text-only differences as a sub-diff.
        try:
            text = Difference.from_operation(Htmltotext, self.path, other.path)
            if text is not None:
                difference.add_details([text])
        except RequiredToolNotFound as exc:  # noqa
            difference.add_comment(exc.get_comment())
        except subprocess.CalledProcessError:
            pass

        return difference



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020, 2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
import time

import pytest

from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.directory import compare_directories
from diffoscope.comparators.utils.compare import compare_root_paths
from diffoscope.comparators.utils.specialize import specialize
from diffoscope.config import Config

from ..utils.data import data, get_data, assert_diff

TEST_FILE1_PATH = data("text_ascii1")
TEST_FILE2_PATH = data("text_ascii2")


def test_no_differences():
    x = os.path.dirname(__file__)
    difference = compare_directories(x, x)
    assert difference is None


def test_no_differences_with_extra_slash():
    x = os.path.dirname(__file__)
    difference = compare_directories(x + "/", x)
    assert difference is None


@pytest.fixture
def differences(tmpdir):
    tmpdir.mkdir("a")
    tmpdir.mkdir("a/dir")
    tmpdir.mkdir("b")
    tmpdir.mkdir("b/dir")
    shutil.copy(TEST_FILE1_PATH, str(tmpdir.join("a/dir/text")))
    shutil.copy(TEST_FILE2_PATH, str(tmpdir.join("b/dir/text")))
    os.utime(str(tmpdir.join("a/dir/text")), (0, 0))
    os.utime(str(tmpdir.join("b/dir/text")), (0, 0))
    os.utime(str(tmpdir.join("a/dir")), (0, 0))
    os.utime(str(tmpdir.join("b/dir")), (0, 0))
    os.utime(str(tmpdir.join("a")), (0, 0))
    os.utime(str(tmpdir.join("b")), (0, 0))
    return compare_directories(
        str(tmpdir.join("a")), str(tmpdir.join("b"))
    ).details


def test_content(differences):
    assert "/dir" in differences[0].source1
    assert "/dir/text" in differences[0].details[0].source1
    expected_diff = get_data("text_ascii_expected_diff")
    assert differences[0].details[0].unified_diff == expected_diff


def test_stat(differences):
    assert "stat" in differences[0].details[0].details[0].source1


def test_compare_to_file(tmpdir):
    path = str(tmpdir.join("file"))

    with open(path, "w") as f:
        f.write("content")

    a = specialize(FilesystemFile(str(tmpdir.mkdir("dir"))))
    b = specialize(FilesystemFile(path))

    assert_diff(a.compare(b), "test_directory_file_diff")


def test_compare_to_device(tmpdir):
    a = specialize(FilesystemFile(str(tmpdir.mkdir("dir"))))
    b = specialize(FilesystemFile("/dev/null"))

    assert_diff(a.compare(b), "test_directory_device_diff")


def test_compare_to_symlink(tmpdir):
    path = str(tmpdir.join("src"))
    os.symlink("/etc/passwd", path)

    a = specialize(FilesystemFile(str(tmpdir.mkdir("dir"))))
    b = specialize(FilesystemFile(path))

    assert_diff(a.compare(b), "test_directory_symlink_diff")


def test_compare_to_dangling_symlink(tmpdir):
    path = str(tmpdir.join("src"))
    os.symlink("/dangling", path)

    a = specialize(FilesystemFile(str(tmpdir.mkdir("dir"))))
    b = specialize(FilesystemFile(path))

    assert_diff(a.compare(b), "test_directory_symlink_diff")


def test_compare_both_ways(tmpdir):
    """
    Comparing a directory with a file shouldn't crash, but nor should as
    comparing a file with a directory either. (Re: #292)
    """

    a = str(tmpdir)
    b = TEST_FILE1_PATH

    assert_diff(compare_root_paths(a, b), "test_directory_a_b_diff")
    assert_diff(compare_root_paths(b, a), "test_directory_b_a_diff")


def test_identical_files_different_mtime(monkeypatch, tmpdir):

    monkeypatch.setattr(Config(), "exclude_directory_metadata", "no")

    file1_path = str(tmpdir.join("file1"))
    file2_path = str(tmpdir.join("file2"))

    now = time.time()
    for path, t in ((file1_path, now - 3600), (file2_path, now)):
        with open(path, "w") as f:
            f.write("identical content")
        os.utime(path, (t, t))

    difference = compare_root_paths(file1_path, file2_path)
    assert difference is not None



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015 Clemens Lang <cal@macports.org>
# Copyright © 2016-2021 Chris Lamb <lamby@debian.org>
# Copyright © 2021 Jean-Romain Garnier <salsa@jean-romain.com>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.config import Config
from diffoscope.comparators.macho import MachoFile
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist
from .test_rlib import llvm_version

obj1 = load_fixture("test1.macho")
obj2 = load_fixture("test2.macho")


@pytest.fixture(scope="function", autouse=True)
def init_tests(request, monkeypatch):
    # Ignore radare2 commands so decompiling is skipped
    # See test_macho_decompiler.py for tests related to decompiler
    monkeypatch.setattr(Config(), "exclude_commands", ["^radare2.*"])


def test_obj_identification(obj1):
    assert isinstance(obj1, MachoFile)


@skip_unless_tools_exist("llvm-readobj", "llvm-objdump")
def test_obj_no_differences(obj1):
    difference = obj1.compare(obj1)
    assert difference is None


@pytest.fixture
def obj_differences(obj1, obj2):
    return obj1.compare(obj2).details


@skip_unless_tools_exist("otool", "lipo")
def test_otool_obj_compare_non_existing(monkeypatch, obj1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = obj1.compare(MissingFile("/nonexisting", obj1))
    assert difference.source2 == "/nonexisting"
    assert len(difference.details) > 0


@skip_unless_tools_exist("otool", "lipo")
def test_otool_diff(obj_differences):
    assert len(obj_differences) == 2
    assert_diff(obj_differences[0], "macho_otool_expected_diff_strings")

    arch_differences = obj_differences[-1].details
    assert len(arch_differences) == 6

    filenames = [
        "macho_otool_expected_diff_headers",
        "macho_otool_expected_diff_libraries",
        "macho_otool_expected_diff_indirect_symbols",
        "macho_otool_expected_diff__text",
        "macho_otool_expected_diff__cstring",
        "macho_otool_expected_diff__data",
    ]
    for idx, diff in enumerate(arch_differences):
        assert_diff(diff, filenames[idx])


@skip_unless_tools_exist("llvm-readobj", "llvm-objdump")
def test_llvm_obj_compare_non_existing(monkeypatch, obj1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = obj1.compare(MissingFile("/nonexisting", obj1))
    assert difference.source2 == "/nonexisting"
    assert len(difference.details) > 0


@skip_unless_tools_exist("llvm-readobj", "llvm-objdump")
def test_llvm_diff(obj_differences):
    if llvm_version() < "13":
        diff_symbols = "macho_llvm_expected_diff_symbols_llvm_11"
    else:
        diff_symbols = "macho_llvm_expected_diff_symbols"

    # Headers
    assert len(obj_differences) == 8
    filenames = [
        "macho_llvm_expected_diff_strings",
        "macho_llvm_expected_diff_file_headers",
        "macho_llvm_expected_diff_needed_libs",
        diff_symbols,
        "macho_llvm_expected_diff_dyn_symbols",
        "macho_llvm_expected_diff_relocations",
        "macho_llvm_expected_diff_dyn_relocations",
    ]
    for idx, diff in enumerate(obj_differences[:-1]):
        assert_diff(diff, filenames[idx])

    # Sections
    arch_differences = obj_differences[-1].details
    assert len(arch_differences) == 7
    if llvm_version() < "16":
        filenames = [
            "macho_llvm_expected_diff__text",
            "macho_llvm_expected_diff__stubs",
            "macho_llvm_expected_diff__stub_helper",
            "macho_llvm_expected_diff__cstring",
            "macho_llvm_expected_diff__unwind_info",
            "macho_llvm_expected_diff__eh_frame",
            "macho_llvm_expected_diff__la_symbol_ptr",
        ]
    else:
        filenames = [
            "macho_llvm_expected_diff__text_16",
            "macho_llvm_expected_diff__stubs_16",
            "macho_llvm_expected_diff__stub_helper_16",
            "macho_llvm_expected_diff__cstring",
            "macho_llvm_expected_diff__unwind_info",
            "macho_llvm_expected_diff__eh_frame",
            "macho_llvm_expected_diff__la_symbol_ptr",
        ]
    for idx, diff in enumerate(arch_differences):
        assert_diff(diff, filenames[idx])



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2017, 2020, 2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import shutil
import subprocess
import pytest

from diffoscope.comparators.xz import XzFile
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least
from ..utils.nonexisting import assert_non_existing

xz1 = load_fixture("test1.xz")
xz2 = load_fixture("test2.xz")
xz3 = load_fixture("test3.xz")
xz4 = load_fixture("test4.xz")


def xz_version():
    try:
        out = subprocess.check_output(["xz", "--version"])
    except subprocess.CalledProcessError as e:
        out = e.output
    return out.decode("UTF-8").split()[3].strip()


def test_identification(xz1):
    assert isinstance(xz1, XzFile)


def test_no_differences(xz1):
    difference = xz1.compare(xz1)
    assert difference is None


@pytest.fixture
def differences(xz1, xz2):
    return xz1.compare(xz2).details


@skip_unless_tools_exist("xz")
def test_content_source(differences):
    assert differences[0].source1 == "test1"
    assert differences[0].source2 == "test2"


@skip_unless_tools_exist("xz")
def test_content_source_without_extension(tmpdir, xz1, xz2):
    path1 = str(tmpdir.join("test1"))
    path2 = str(tmpdir.join("test2"))
    shutil.copy(xz1.path, path1)
    shutil.copy(xz2.path, path2)
    xz1 = specialize(FilesystemFile(path1))
    xz2 = specialize(FilesystemFile(path2))
    difference = xz1.compare(xz2).details
    assert difference[0].source1 == "test1-content"
    assert difference[0].source2 == "test2-content"


@skip_unless_tools_exist("xz")
def test_content_diff(differences):
    assert_diff(differences[0], "text_ascii_expected_diff")


@skip_unless_tools_exist("xz")
def test_compare_non_existing(monkeypatch, xz1):
    assert_non_existing(monkeypatch, xz1)


@pytest.fixture
def differences_verbose(xz3, xz4):
    return xz3.compare(xz4).details


@skip_unless_tools_exist("xz")
@skip_unless_tool_is_at_least("xz", xz_version, "5.4.1")
def test_content_source_verbose(differences_verbose):
    assert differences_verbose[0].source1 == "xz --list"
    assert differences_verbose[0].source2 == "xz --list"
    assert_diff(differences_verbose[0], "text_xz_list")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2018-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command


class DbDump(Command):
    @tool_required("db_dump")
    def cmdline(self):
        return ("db_dump", "-d", "a", self.path)

    def filter(self, line):
        val = line.decode("utf-8")
        # We must strip some fields as libdb itself does not repeatedly read
        # its own metadata reliably, even on the same file.
        for x in ("h_hash: ", "bt_compare: ", "\tuid: "):
            if val.startswith(x):
                return b""
        return line


class BerkeleyDBFile(File):
    DESCRIPTION = "Berkeley DB database files"
    FILE_TYPE_RE = re.compile(r"^Berkeley DB ")

    def compare_details(self, other, source=None):
        return [
            Difference.from_operation(
                DbDump, self.path, other.path, source="Berkeley DB file"
            )
        ]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017 Juliana Rodrigues <juliana.orod@gmail.com>
# Copyright © 2017-2022, 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import pyexpat
import logging

from xml.parsers.expat import ExpatError

from diffoscope.comparators.utils.file import File
from diffoscope.difference import Difference
from diffoscope.tools import python_module_missing

from .missing_file import MissingFile

try:
    defusedxml = True
    from defusedxml import minidom
except ImportError:
    defusedxml = False
    python_module_missing("defusedxml")
    from xml.dom import minidom


def is_vulnerable_xml_parser():
    # We assume defusedxml is safe
    if defusedxml:
        return False

    """
    As described in Pythons module documentation versions of expat <= 2.4.1
    (released on 2021-05-23) are vulnerable to exponential/quadratic entity
    expansion while versions <= 2.6.0 (released on 2024-02-06) are vulnerable to
    large tokens. Since expat is usually provided by the system and not
    directly bundled with Python, even recent Python installations can still be
    vulnerable due to old expat versions.

        <https://salsa.debian.org/reproducible-builds/diffoscope/-/issues/397>
    """

    pyexpat_version = tuple(
        int(x) for x in pyexpat.EXPAT_VERSION.split("_", 1)[1].split(".")
    )

    return pyexpat_version < (2, 6, 0)


logger = logging.getLogger(__name__)
if is_vulnerable_xml_parser():
    logger.warning(
        "Vulnerable version of pyexpat detected; disabling comparison of XML documents. Install defusedxml or upgrade your pyexpat."
    )


def _format(node):
    """
    Removes *inplace* spaces from minidom.Document

    Args:
        node -- A xml.dom.minidom.Document object

    Returns:
        void
    """
    for n in node.childNodes:
        if n.nodeType == n.TEXT_NODE:
            if n.nodeValue:
                n.nodeValue = n.nodeValue.strip()
        elif n.nodeType == n.ELEMENT_NODE:
            _format(n)


def _parse(file):
    """
    Formats a minidom.Document file and returns XML as string.

    Args:
        file -- An io.TextIOWrapper object

    Returns:
        str: formated string object
    """

    if defusedxml:
        xml = minidom.parse(file, forbid_entities=False, forbid_external=False)
    else:
        xml = minidom.parse(file)

    _format(xml)
    xml.normalize()

    return xml.toprettyxml(indent=2 * " ", encoding="utf-8").decode("utf-8")


class XMLFile(File):
    """
    XML Files Comparison class

    Attributes:
        FILE_EXTENSION_SUFFIX (str): xml file extension suffix
    """

    DESCRIPTION = "XML files"
    FILE_TYPE_RE = re.compile(
        r"^(XML \S+ document|GnuCash file|XHTML document|SVG Scalable Vector Graphics image)"
    )

    @classmethod
    def recognizes(cls, file):
        """
        Identifies if a given file has XML extension

        Args:
            file - a diffoscope.comparators.utils.file.File object

        Returns:
            False if file is not a XML File, True otherwise
        """

        # Emulate FALLBACK_FILE_EXTENSION_SUFFIX = {".xml"}
        if not super().recognizes(file) and not file.name.endswith(".xml"):
            return False

        # Check that we aren't about to open an untrusted file with an old, and
        # potentially vulnerable, version of pyexpat.
        if is_vulnerable_xml_parser():
            return False

        with open(file.path) as f:
            try:
                file.parsed = _parse(f)
            except (ExpatError, UnicodeDecodeError):
                return False

        return True

    def compare_details(self, other, source=None):
        """
        Compares self.object with another, returning a Difference object

        Args:
            other    -- A XMLFile object
            source

        Returns:
            A diffoscope.difference.Difference object
        """
        if isinstance(other, MissingFile):
            return [
                Difference(
                    self.name,
                    other.name,
                    comment="Trying to compare two non-existing files.",
                )
            ]

        difference = Difference.from_text(
            self.dumps(self), self.dumps(other), self.name, other.name
        )

        if difference:
            difference.check_for_ordering_differences()

        return [difference]

    def dumps(self, file):
        """
        Opens a XMLFile and returns its parsed content

        Args:
            file -- XMLFile object

        Returns:
            str -- Formatted XML content from file
        """
        if file.parsed:
            return file.parsed

        with open(file.path) as f:
            return _parse(f)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.sphinx import SphinxInventoryFile

from ..utils.data import assert_diff, load_fixture

inv1 = load_fixture("test1.inv")
inv2 = load_fixture("test2.inv")


def test_identification(inv1, inv2):
    assert isinstance(inv1, SphinxInventoryFile)
    assert isinstance(inv2, SphinxInventoryFile)


def test_no_differences(inv1):
    assert inv1.compare(inv1) is None


@pytest.fixture
def differences(inv1, inv2):
    return inv1.compare(inv2).details


def test_diff(differences):
    assert_diff(differences[0], "inv_expected_diff")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

import diffoscope.comparators

from diffoscope.config import Config
from diffoscope.comparators.deb import DebFile, Md5sumsFile, DebDataTarFile
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.missing_file import MissingFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.tools import (
    skip_unless_tools_exist,
    skip_unless_file_version_is_at_least,
)
from ..utils.data import load_fixture, get_data

deb1 = load_fixture("test1.deb")
deb2 = load_fixture("test2.deb")


def test_identification(deb1):
    assert isinstance(deb1, DebFile)


def test_no_differences(deb1):
    difference = deb1.compare(deb1)
    assert difference is None


@pytest.fixture
def differences(deb1, deb2):
    return deb1.compare(deb2).details


def test_metadata(differences):
    expected_diff = get_data("deb_metadata_expected_diff")
    assert differences[0].unified_diff == expected_diff


def test_compressed_files(differences):
    assert differences[1].source1 == "control.tar.gz"
    assert differences[2].source1 == "data.tar.gz"


def test_identification_of_md5sums_outside_deb(tmpdir):
    path = str(tmpdir.join("md5sums"))
    open(path, "w")
    f = specialize(FilesystemFile(path))
    assert type(f) is FilesystemFile


def test_identification_of_md5sums_in_deb(deb1, deb2, monkeypatch):
    orig_func = Md5sumsFile.recognizes

    @staticmethod
    def probe(file):
        ret = orig_func(file)
        if ret:
            test_identification_of_md5sums_in_deb.found = True
        return ret

    test_identification_of_md5sums_in_deb.found = False
    monkeypatch.setattr(Md5sumsFile, "recognizes", probe)
    deb1.compare(deb2)
    assert test_identification_of_md5sums_in_deb.found


def test_md5sums(differences):
    assert differences[1].details[0].details[1].details[0].comment
    # we skip diffing the md5sums since it's not interesting, the actual files are diffed instead
    assert not differences[1].details[0].details[1].details[0].unified_diff
    assert not differences[1].details[0].details[1].details[0].details


def test_identical_files_in_md5sums(deb1, deb2):
    for name in [
        "./usr/share/doc/test/README.Debian",
        "./usr/share/doc/test/copyright",
    ]:
        assert deb1.md5sums[name] == deb2.md5sums[name]


def test_identification_of_data_tar(deb1, deb2, monkeypatch):
    orig_func = DebDataTarFile.recognizes

    @staticmethod
    def probe(file):
        ret = orig_func(file)
        if ret:
            test_identification_of_data_tar.found = True
        return ret

    test_identification_of_data_tar.found = False
    monkeypatch.setattr(DebDataTarFile, "recognizes", probe)
    deb1.compare(deb2)
    assert test_identification_of_data_tar.found


def test_skip_comparison_of_known_identical_files(deb1, deb2, monkeypatch):
    compared = set()
    orig_func = diffoscope.comparators.utils.compare.compare_files

    def probe(file1, file2, **kwargs):
        compared.add(file1.name)
        return orig_func(file1, file2, **kwargs)

    monkeypatch.setattr(
        diffoscope.comparators.utils.compare, "compare_files", probe
    )
    deb1.compare(deb2)
    assert "./usr/share/doc/test/README.Debian" not in compared


def test_compare_non_existing(monkeypatch, deb1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = deb1.compare(MissingFile("/nonexisting", deb1))
    assert difference.source2 == "/nonexisting"
    assert difference.details[-1].source2 == "/dev/null"


bug881937_deb1 = load_fixture("bug881937_1.deb")
bug881937_deb2 = load_fixture("bug881937_2.deb")
bug903391_deb1 = load_fixture("bug903391_1.deb")
bug903391_deb2 = load_fixture("bug903391_2.deb")
bug903401_deb1 = load_fixture("bug903401_1.deb")
bug903401_deb2 = load_fixture("bug903401_2.deb")
bug903565_deb1 = load_fixture("bug903565_1.deb")
bug903565_deb2 = load_fixture("bug903565_2.deb")


@skip_unless_tools_exist("xz")
@skip_unless_file_version_is_at_least("5.37")
def test_compare_different_compression(bug881937_deb1, bug881937_deb2):
    difference = bug881937_deb1.compare(bug881937_deb2)
    assert difference.details[2].source1 == "control.tar.gz"
    assert difference.details[2].source2 == "control.tar.xz"
    expected_diff = get_data("bug881937_control_expected_diff")
    assert (
        difference.details[2].details[2].details[1].unified_diff
        == expected_diff
    )


def test_uncompressed_data_tar(bug903401_deb1, bug903401_deb2):
    bug903401_deb1.compare(bug903401_deb2)


def test_uncompressed_control_tar(bug903391_deb1, bug903391_deb2):
    bug903391_deb1.compare(bug903391_deb2)


def test_compare_different_compression_multiple_files(
    bug903565_deb1, bug903565_deb2
):
    bug903565_deb1.compare(bug903565_deb2)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2017, 2019-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators import ComparatorManager
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import (
    load_fixture,
    data,
    get_data,
    normalize_zeros,
    assert_diff,
)
from ..utils.tools import skip_unless_tools_exist, skip_unless_module_exists
from ..utils.nonexisting import assert_non_existing

try:
    from diffoscope.comparators.rpm import RpmFile
except ImportError:
    from diffoscope.comparators.rpm_fallback import RpmFile

rpm1 = load_fixture("test1.rpm")
rpm2 = load_fixture("test2.rpm")


def test_identification(rpm1):
    assert isinstance(rpm1, RpmFile)


@skip_unless_module_exists("rpm")
def test_no_differences(rpm1):
    difference = rpm1.compare(rpm1)
    assert difference is None


@pytest.fixture
def differences(rpm1, rpm2):
    return rpm1.compare(rpm2).details


@skip_unless_module_exists("rpm")
@skip_unless_tools_exist("rpm2cpio")
def test_header(differences):
    assert differences[0].source1 == "header"
    assert_diff(differences[0], "rpm_header_expected_diff")


@skip_unless_module_exists("rpm")
@skip_unless_tools_exist("rpm2cpio")
def test_listing(differences):
    assert differences[1].source1 == "content"
    assert differences[1].details[0].source1 == "file list"
    assert_diff(differences[1].details[0], "rpm_listing_expected_diff")


@skip_unless_module_exists("rpm")
@skip_unless_tools_exist("rpm2cpio")
def test_content(differences):
    assert differences[1].source1 == "content"
    assert differences[1].details[1].source1 == "./dir/text"
    assert_diff(differences[1].details[1], "text_ascii_expected_diff")


@skip_unless_module_exists("rpm")
@skip_unless_tools_exist("rpm2cpio")
def test_compare_non_existing(monkeypatch, rpm1):
    assert_non_existing(monkeypatch, rpm1)


@skip_unless_tools_exist("xxd")
def test_fallback_comparison(monkeypatch):
    manager = ComparatorManager()
    monkeypatch.setattr(manager, "COMPARATORS", (("rpm_fallback.RpmFile",),))
    manager.reload()

    # Re-specialize after reloading our Comparators
    rpm1 = specialize(FilesystemFile(data("test1.rpm")))
    rpm2 = specialize(FilesystemFile(data("test2.rpm")))

    assert rpm1.compare(rpm1) is None
    assert rpm2.compare(rpm2) is None

    expected_diff = get_data("rpm_fallback_expected_diff")
    assert normalize_zeros(rpm1.compare(rpm2).unified_diff) == expected_diff



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020, 2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import os.path
import logging
import subprocess

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.archive import Archive
from .utils.command import Command

logger = logging.getLogger(__name__)


class XZList(Command):
    @tool_required("xz")
    def cmdline(self):
        return ("xz", "--list", "--verbose", "--verbose", self.path)

    def filter(self, line):
        # Strip the first line
        if line.startswith(self.path.encode("utf-8")):
            return b""
        return line


class XzContainer(Archive):
    def open_archive(self):
        return self

    def close_archive(self):
        pass

    def get_member_names(self):
        return [self.get_compressed_content_name(".xz")]

    @tool_required("xz")
    def extract(self, member_name, dest_dir):
        dest_path = os.path.join(dest_dir, member_name)
        logger.debug("xz extracting to %s", dest_path)
        with open(dest_path, "wb") as fp:
            subprocess.check_call(
                ["xz", "--decompress", "--stdout", self.source.path],
                stdout=fp,
                stderr=None,
            )
        return dest_path


class XzFile(File):
    DESCRIPTION = "XZ compressed files"
    CONTAINER_CLASSES = [XzContainer]
    FILE_TYPE_RE = re.compile(r"^XZ compressed data\b")

    # Work around file(1) Debian bug #876316
    FALLBACK_FILE_EXTENSION_SUFFIX = {".xz"}
    FALLBACK_FILE_TYPE_HEADER_PREFIX = b"\xfd7zXZ\x00"

    def compare(self, other, source=None):
        difference = super().compare(other, source)

        # Append xz --list *after* showing any container differences.
        if difference and not difference.details and isinstance(other, XzFile):
            xz_list = Difference.from_operation(
                XZList, self.path, other.path, source="xz --list"
            )
            if xz_list:
                difference.add_details([xz_list])

        return difference



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import pytest

from diffoscope.main import main


def run(capsys, *args):
    with pytest.raises(SystemExit) as exc:
        main(
            args
            + tuple(
                os.path.join(os.path.dirname(__file__), "data", x)
                for x in ("test1.tar", "test2.tar")
            )
        )

    out, err = capsys.readouterr()

    assert err == ""

    return exc.value.code, out


def test_none(capsys):
    ret, out = run(capsys)
    # Make sure the output doesn't contain any [masked]
    assert re.search(r"\[masked\]", out) is None
    assert ret == 1


def test_all(capsys):
    ret, out = run(capsys, "--diff-mask=.*")

    # Make sure the correct sections were masked
    assert "file list" not in out
    assert "dir/link" not in out

    # Make sure the output contains only [masked]
    # Lines of content start with "│ ", and then either have a +, a - or a space
    # depending on the type of change
    # It should then only contain "[masked]" until the end of the string
    assert re.search(r"│\s[\s\+\-](?!(\[masked\])+)", out) is None
    assert ret == 1


def test_specific(capsys):
    ret, out = run(capsys, "--diff-mask=^Lorem")
    # Make sure only the Lorem ipsum at the start of the line was masked
    assert "[masked] ipsum dolor sit amet" in out
    assert '"Lorem ipsum"' in out
    assert ret == 1



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from ..utils.data import load_fixture

archive1 = load_fixture("archive1.tar")
archive2 = load_fixture("archive2.tar")


@pytest.fixture
def differences(archive1, archive2):
    return archive1.compare(archive2).details


def test_compressed_content_name(differences):
    assert differences[1].details[1].source1 == "compressed"
    assert differences[1].details[1].source2 == "compressed"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2020, 2024-2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.deb import DebFile
from diffoscope.comparators.zip import ZipFile
from diffoscope.comparators.gzip import GzipFile

from .utils.data import load_fixture, assert_diff
from .utils.tools import (
    skip_unless_file_version_is_at_least,
    file_version_is_ge,
)

quine1 = load_fixture("quine.gz")
quine2 = load_fixture("quine.zip")
quine3 = load_fixture("quine_a.deb")
quine4 = load_fixture("quine_b.deb")

"""
Check that we are not recursively unpacking the quines in an infinite loop. See
<https://research.swtch.com/zip> and <https://bugs.debian.org/780761>.
"""


def test_identification(quine1, quine2):
    assert isinstance(quine1, GzipFile)
    assert isinstance(quine2, ZipFile)


def test_no_differences(quine1):
    difference = quine1.compare(quine1)
    assert difference is None


@pytest.fixture
def differences(quine1, quine2):
    return quine1.compare(quine2).details


@skip_unless_file_version_is_at_least("5.47")
def test_difference(differences):
    assert_diff(differences[0], "quine_expected_diff")


def test_identification_deb(quine3, quine4):
    assert isinstance(quine3, DebFile)
    assert isinstance(quine4, DebFile)


@pytest.fixture
def differences_deb(quine3, quine4):
    return quine3.compare(quine4).details


def test_differences_deb(differences_deb):
    assert_diff(differences_deb[0], "quine_deb_expected_diff")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.config import Config
from diffoscope.comparators.asar import AsarFile
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing

asar1 = load_fixture("test1.asar")
asar2 = load_fixture("test2.asar")


def test_identification(asar1):
    assert isinstance(asar1, AsarFile)


def test_no_differences(asar1):
    difference = asar1.compare(asar1)
    assert difference is None


@pytest.fixture
def differences(asar1, asar2):
    return asar1.compare(asar2).details


@skip_unless_tools_exist("asar")
def test_text_file(differences):
    assert differences[0].source1 == "dir/text"
    assert differences[0].source2 == "dir/text"
    expected_diff = get_data("text_ascii_expected_diff")
    assert differences[0].unified_diff == expected_diff



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2020, 2022, 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import pytest
import signal
import tempfile

from diffoscope.main import main

TEST_TAR1_PATH = os.path.join(os.path.dirname(__file__), "data/test1.tar")
TEST_TAR2_PATH = os.path.join(os.path.dirname(__file__), "data/test2.tar")
TEST_TARS = (TEST_TAR1_PATH, TEST_TAR2_PATH)


def run(capsys, *args):
    with pytest.raises(SystemExit) as exc:
        main(args)

    out, err = capsys.readouterr()

    return exc.value.code, out, err


def test_non_existing_files(capsys):
    ret, _, err = run(capsys, "/nonexisting1", "/nonexisting2")

    assert ret == 2
    assert "/nonexisting1: No such file or directory" in err
    assert "/nonexisting2: No such file or directory" in err


def test_non_existing_left_with_new_file(capsys):
    ret, out, _ = run(capsys, "--new-file", "/nonexisting1", __file__)

    assert ret == 1
    assert "--- /nonexisting1" in out
    assert ("+++ %s" % __file__) in out


def test_non_existing_right_with_new_file(capsys):
    ret, out, _ = run(capsys, "--new-file", __file__, "/nonexisting2")

    assert ret == 1
    assert ("--- %s" % __file__) in out
    assert "+++ /nonexisting2" in out


def test_non_existing_files_with_new_file(capsys):
    ret, out, _ = run(capsys, "--new-file", "/nonexisting1", "/nonexisting2")

    assert ret == 1
    assert "--- /nonexisting1" in out
    assert "+++ /nonexisting2" in out
    assert "Trying to compare two non-existing files." in out


def test_remove_temp_files_on_sigterm(capsys, tmpdir, monkeypatch):
    pid = os.fork()

    if pid == 0:

        def suicide(*args):
            os.kill(os.getpid(), signal.SIGTERM)

        monkeypatch.setattr(
            "diffoscope.comparators.text.TextFile.compare", suicide
        )
        tempfile.tempdir = str(tmpdir)

        ret, _, _ = run(capsys, *TEST_TARS)
        os._exit(ret)
    else:
        _, ret = os.waitpid(pid, 0)
        assert (ret >> 8) == 2  # having received SIGTERM is trouble
        assert os.listdir(str(tmpdir)) == []


def test_ctrl_c_handling(tmpdir, monkeypatch, capsys):
    monkeypatch.setattr("tempfile.tempdir", str(tmpdir))

    def interrupt(*args):
        raise KeyboardInterrupt

    monkeypatch.setattr(
        "diffoscope.comparators.text.TextFile.compare", interrupt
    )

    ret, _, err = run(capsys, *TEST_TARS)

    assert "" in err
    assert ret == 2
    assert os.listdir(str(tmpdir)) == []


def test_no_differences(capsys):
    ret, out, err = run(capsys, TEST_TAR1_PATH, TEST_TAR1_PATH)

    assert ret == 0
    assert err == ""
    assert out == ""


def test_no_differences_directories(capsys, tmpdir):
    def create_dir(x):
        path = str(tmpdir.mkdir(x))
        os.utime(path, (0, 0))  # Ensure consistent mtime
        return path

    ret, out, err = run(capsys, create_dir("a"), create_dir("b"))

    assert ret == 0
    # if getfacl is not available there will be a warning message on stderr
    # if it's available then err should be empty
    assert err == "" or "getfacl" in err
    assert out == ""


def test_list_tools(capsys):
    ret, out, err = run(capsys, "--list-tools")

    assert ret == 0
    assert err == ""
    assert "External-Tools-Required: " in out
    assert "xxd," in out


def test_list_missing_tools(capsys):
    ret, out, err = run(capsys, "--list-missing-tools")

    assert ret == 0
    assert err == ""
    assert "External-Tools-Required: " in out
    # No assertions about the contents of the output since we don't control
    # what's installed in the test environment


def test_profiling(capsys):
    ret, out, err = run(capsys, TEST_TAR1_PATH, TEST_TAR1_PATH, "--profile=-")

    assert ret == 0
    assert "Profiling output for" in out
    assert err == ""


def test_non_unicode_filename(capsys, tmpdir):
    # Bug reference: https://bugs.debian.org/898022
    path = str(tmpdir.dirpath()).encode("utf-8")
    a = os.path.join(path, b"\xf0\x28\x8c\x28")
    b = os.path.join(path, b"\xf0\x28\x8c\x29")
    with open(a, "w"), open(b, "w"):
        pass

    # sys.argv does pretty much this decoding to arguments
    files = [x.decode("utf-8", errors="surrogateescape") for x in (a, b)]
    ret, out, err = run(capsys, *files)

    assert ret == 0
    assert out == err == ""


def test_help(capsys, monkeypatch):
    # Fake --help in sys.argv so that we include out file formats in the output
    monkeypatch.setattr(sys, "argv", ["diffoscope", "--help"])

    ret, out, err = run(capsys, "--help")

    assert err == ""
    assert "file formats supported:" in out


def test_usage(capsys):
    ret, out, err = run(capsys, "--usage")

    assert out == ""
    assert "usage:" in err


def test_valid_css(capsys):
    ret, out, err = run(
        capsys,
        "--css",
        "http://example.tld/file.css",
        TEST_TAR1_PATH,
        TEST_TAR1_PATH,
    )

    assert ret == 0
    assert err == ""
    assert out == ""


def test_invalid_css(capsys):
    ret, out, err = run(
        capsys,
        "--css",
        '"><svg/onload=alert(43433)>',
        TEST_TAR1_PATH,
        TEST_TAR1_PATH,
    )

    assert ret == 2
    assert "is not a valid CSS" in err
    assert out == ""



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020, 2023, 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.text import TextFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import data, load_fixture, assert_diff
from ..utils.nonexisting import assert_non_existing

ascii1 = load_fixture("text_ascii1")
ascii2 = load_fixture("text_ascii2")


def test_no_differences(ascii1):
    difference = ascii1.compare(ascii1)
    assert difference is None


def test_difference_in_ascii(ascii1, ascii2):
    difference = ascii1.compare(ascii2)
    assert difference is not None
    assert_diff(difference, "text_ascii_expected_diff")
    assert not difference.comments
    assert len(difference.details) == 0


unicode1 = load_fixture("text_unicode1")
unicode2 = load_fixture("text_unicode2")


def test_difference_in_unicode(unicode1, unicode2):
    difference = unicode1.compare(unicode2)
    expected_diff = open(
        data("text_unicode_expected_diff"), encoding="utf-8"
    ).read()
    assert difference.unified_diff == expected_diff


iso8859 = load_fixture("text_iso8859")


def test_difference_between_iso88591_and_unicode(iso8859, unicode1):
    difference = iso8859.compare(unicode1)
    expected_diff = open(
        data("text_iso8859_expected_diff"), encoding="utf-8"
    ).read()
    assert difference.unified_diff == expected_diff


def test_difference_between_iso88591_and_unicode_only(iso8859, tmpdir):
    utf8_path = str(tmpdir.join("utf8"))
    with open(utf8_path, "wb") as f:
        f.write(
            open(data("text_iso8859"), encoding="iso8859-1")
            .read()
            .encode("utf-8")
        )
    utf8 = specialize(FilesystemFile(utf8_path))
    difference = iso8859.compare(utf8)
    assert difference.unified_diff is None
    assert difference.details[0].source1 == "encoding"


def test_compare_non_existing(monkeypatch, ascii1):
    assert_non_existing(
        monkeypatch, ascii1, has_null_source=False, has_details=False
    )


text_order1 = load_fixture("text_order1")
text_order2 = load_fixture("text_order2")
text_endings1 = load_fixture("text_endings1")
text_endings2 = load_fixture("text_endings2")


def test_ordering_differences(text_order1, text_order2):
    difference = text_order1.compare(text_order2)
    assert difference.comments == ["Ordering differences only"]
    assert_diff(difference, "text_order_expected_diff")


def test_ending_differences(text_endings1, text_endings2):
    difference = text_endings1.compare(text_endings2)
    assert difference.comments == ["Line-ending differences only"]
    assert_diff(difference, "text_endings_expected_diff")


def test_text_fallback(tmp_path):
    """
    If we don't know the exact type but it looks like text, compare it as text.
    """
    temp = tmp_path / "temp.msc"
    temp.write_text("msc{a, b;}")
    specialized = specialize(FilesystemFile(str(temp)))
    assert isinstance(specialized, TextFile)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020, 2022, 2024-2026 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.comparators.zip import (
    ZipFile,
    MozillaZipFile,
    NuGetPackageFile,
    JmodJavaModule,
    zipdetails_version,
)

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least
from ..utils.nonexisting import assert_non_existing

zip1 = load_fixture("test1.zip")
zip2 = load_fixture("test2.zip")
zip3 = load_fixture("test3.zip")
encrypted_zip1 = load_fixture("encrypted1.zip")
encrypted_zip2 = load_fixture("encrypted2.zip")
mozzip1 = load_fixture("test1.mozzip")
mozzip2 = load_fixture("test2.mozzip")
nupkg1 = load_fixture("test1.nupkg")
nupkg2 = load_fixture("test2.nupkg")
jmod1 = load_fixture("test1.jmod")
jmod2 = load_fixture("test2.jmod")
test_comment1 = load_fixture("test_comment1.zip")
test_comment2 = load_fixture("test_comment2.zip")


def io_compress_zip_version():
    try:
        return subprocess.check_output(
            [
                "perl",
                "-MIO::Compress::Zip",
                "-e",
                "print $IO::Compress::Zip::VERSION",
            ]
        ).decode("UTF-8")
    except subprocess.CalledProcessError:
        return "-1"


def test_identification(zip1):
    assert isinstance(zip1, ZipFile)


def test_no_differences(zip1):
    difference = zip1.compare(zip1)
    assert difference is None


@pytest.fixture
def differences(zip1, zip2):
    return zip1.compare(zip2).details


@pytest.fixture
def differences2(zip1, zip3):
    return zip1.compare(zip3).details


@skip_unless_tools_exist("zipinfo", "zipdetails")
@skip_unless_tool_is_at_least("zipdetails", zipdetails_version, "4.008")
@skip_unless_tool_is_at_least("perl", io_compress_zip_version, "2.212")
def test_metadata(differences):
    assert_diff(differences[0], "zip_zipinfo_expected_diff")
    assert_diff(differences[1], "zip_zipdetails_expected_diff")


@skip_unless_tools_exist("zipinfo", "zipdetails")
def test_compressed_files(differences):
    assert differences[2].source1 == "dir/text"
    assert differences[2].source2 == "dir/text"
    assert_diff(differences[2], "text_ascii_expected_diff")


@skip_unless_tools_exist("zipinfo", "bsdtar", "zipdetails")
@skip_unless_tool_is_at_least("zipdetails", zipdetails_version, "4.008")
@skip_unless_tool_is_at_least("perl", io_compress_zip_version, "2.212")
def test_extra_fields(differences2):
    assert_diff(differences2[0], "zip_bsdtar_expected_diff")
    assert_diff(differences2[1], "zip2_zipdetails_expected_diff")


@skip_unless_tools_exist("zipinfo")
def test_compare_non_existing(monkeypatch, zip1):
    assert_non_existing(monkeypatch, zip1)


def test_mozzip_identification(mozzip1):
    assert isinstance(mozzip1, MozillaZipFile)


def test_mozzip_no_differences(mozzip1):
    difference = mozzip1.compare(mozzip1)
    assert difference is None


@pytest.fixture
def mozzip_differences(mozzip1, mozzip2):
    return mozzip1.compare(mozzip2).details


@skip_unless_tools_exist("zipinfo")
def test_mozzip_metadata(mozzip_differences, mozzip1, mozzip2):
    assert_diff(mozzip_differences[0], "mozzip_zipinfo_expected_diff")


@skip_unless_tools_exist("zipinfo")
def test_mozzip_compressed_files(mozzip_differences):
    assert mozzip_differences[-1].source1 == "dir/text"
    assert mozzip_differences[-1].source2 == "dir/text"
    assert_diff(mozzip_differences[-1], "text_ascii_expected_diff")


@skip_unless_tools_exist("zipinfo")
def test_mozzip_compare_non_existing(monkeypatch, mozzip1):
    assert_non_existing(monkeypatch, mozzip1)


def test_nupkg_identification(nupkg1):
    assert isinstance(nupkg1, NuGetPackageFile)


def test_nupkg_no_differences(nupkg1):
    difference = nupkg1.compare(nupkg1)
    assert difference is None


@pytest.fixture
def nupkg_differences(nupkg1, nupkg2):
    return nupkg1.compare(nupkg2).details


@skip_unless_tools_exist("zipinfo")
def test_nupkg_metadata(nupkg_differences, nupkg1, nupkg2):
    assert_diff(nupkg_differences[0], "nupkg_expected_diff")


@skip_unless_tools_exist("zipinfo")
def test_nupkg_compressed_files(nupkg_differences):
    assert (
        nupkg_differences[-1].source1
        == "package/services/metadata/core-properties/b44ebb537bbf4983b9527f9e3820fda6.psmdcp"
    )
    assert (
        nupkg_differences[-1].source2
        == "package/services/metadata/core-properties/08f1f9d8789a4668a128f78560bd0107.psmdcp"
    )


@skip_unless_tools_exist("zipinfo")
def test_nupkg_compare_non_existing(monkeypatch, nupkg1):
    assert_non_existing(monkeypatch, nupkg1)


def test_jmod_identification(jmod1):
    assert isinstance(jmod1, JmodJavaModule)


def test_jmod_no_differences(jmod1):
    difference = jmod1.compare(jmod1)
    assert difference is None


@pytest.fixture
def jmod_differences(jmod1, jmod2):
    return jmod1.compare(jmod2).details


@skip_unless_tools_exist("zipinfo", "zipnote")
@skip_unless_tool_is_at_least("zipdetails", zipdetails_version, "4.008")
@skip_unless_tool_is_at_least("perl", io_compress_zip_version, "2.212")
def test_jmod_metadata(jmod_differences, jmod1, jmod2):
    assert jmod_differences[0].source1 == "zipinfo {}"
    assert jmod_differences[1].source1.startswith("zipnote")
    assert jmod_differences[2].source1.startswith("zipdetails")
    assert_diff(jmod_differences[0], "jmod_zipinfo_expected_diff")
    assert_diff(jmod_differences[2], "jmod_zipdetails_expected_diff")


def test_encrypted(encrypted_zip1, encrypted_zip2):
    difference = encrypted_zip1.compare(encrypted_zip2)
    assert difference is not None


@pytest.fixture
def comment_differences(test_comment1, test_comment2):
    return test_comment1.compare(test_comment2).details


@skip_unless_tools_exist("zipnote", "zipdetails")
@skip_unless_tool_is_at_least("zipdetails", zipdetails_version, "4.008")
def test_commented(comment_differences):
    assert_diff(comment_differences[1], "comment_zipinfo_expected_diff")
    assert_diff(comment_differences[2], "comment_zipdetails_expected_diff")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2021-2022, 2024 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.python import PycFile

from ..utils.data import assert_diff_startswith, load_fixture
from ..utils.nonexisting import assert_non_existing
from ..utils.tools import skip_unless_file_version_is_at_least

pyc1 = load_fixture("test1.pyc-renamed")
pyc2 = load_fixture("test2.pyc-renamed")


@skip_unless_file_version_is_at_least("5.39")
def test_identification(pyc1, pyc2):
    assert isinstance(pyc1, PycFile)
    assert isinstance(pyc2, PycFile)


def test_no_differences(pyc1):
    # Disassembling bytecode prior to Python 3.10 is stable when applied to
    # itself, otherwise various memory offsets (or memory addresses?) are
    # non-deterministic.
    assert pyc1.compare(pyc1) is None


@pytest.fixture
def differences(pyc1, pyc2):
    return pyc1.compare(pyc2).details


@skip_unless_file_version_is_at_least("5.39")
def test_diff(differences):
    assert_diff_startswith(differences[0], "pyc_expected_diff")


def test_compare_non_existing(monkeypatch, pyc1):
    assert_non_existing(
        monkeypatch, pyc1, has_details=False, has_null_source=False
    )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2019-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

HEADER = """<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="utf-8" />
  <meta http-equiv="x-ua-compatible" content="IE=edge">
  <meta name="referrer" content="no-referrer" />
  <meta name="generator" content="diffoscope" />
  <link rel="icon" type="image/png" href="%(favicon)s" />
  <title>%(title)s</title>
%(css_style)s%(css_link)s</head>
<body class="diffoscope">
"""

FOOTER = """<div class="footer">
Generated by
<a href="https://diffoscope.org" rel="noopener noreferrer" target="_blank">
diffoscope</a> %(version)s
</div>
</body>
</html>
"""

STYLES = """body.diffoscope {
  background: white;
  color: black;
}
.diffoscope .footer {
  font-size: small;
}
.diffoscope .difference {
  border: outset #888 1px;
  background: #E8E8E8;
  background: rgba(0,0,0,.1);
  padding: 0.5em;
  margin: 0.5em 0;
}
.diffoscope .difference table {
  table-layout: fixed;
  width: 100%;
  border: 0;
}
.diffoscope .difference th,
.diffoscope .difference td {
  border: 0;
}
.diffoscope table.diff {
  border: 0;
  border-collapse:collapse;
  font-size:0.75em;
  font-family: 'Lucida Console', monospace;
  word-break: break-word;
}
.diffoscope table.diff tr:hover td {
  background: #FFFF00;
}
.diffoscope .line {
  color:#8080a0
}
.diffoscope th {
  background: black;
  color: white
}
.diffoscope .diffunmodified td {
  background: #D0D0E0
}
.diffoscope .diffhunk td {
  background: #A0A0A0
}
.diffoscope .diffadded td {
  background: #CCFFCC
}
.diffoscope .diffdeleted td {
  background: #FFCCCC
}
.diffoscope .diffchanged td {
  background: #FFFFA0
}
.diffoscope ins, del {
  background: #E0C880;
  text-decoration: none
}
.diffoscope .dp {
  color: #B08080
}
.diffoscope .comment {
  font-style: italic;
}
.diffoscope .comment.multiline {
  font-style: normal;
  font-family: monospace;
  white-space: pre;
}
.diffoscope .source {
  font-weight: bold;
}
.diffoscope .error {
  border: solid black 1px;
  background: red;
  color: white;
  padding: 0.2em;
}
.diffoscope .anchor {
  margin-left: 0.5em;
  font-size: 80%;
  color: #333;
  text-decoration: none;
  display: none;
}
.diffoscope .diffheader:hover .anchor {
  display: inline;
}
.diffoscope .diffcontrol, .diffoscope .diffcontrol-nochildren {
  float: left;
  margin-right: 0.3em;
  cursor: pointer;
  display: none; /* currently, only available in html-dir output where jquery is enabled */
}
.diffoscope .colines {
  width: 3em;
}
.diffoscope .coldiff {
  width: 99%;
}
.diffoscope .diffsize {
  float: right;
}
.diffoscope table.diff tr.ondemand td, .diffoscope div.ondemand-details {
  background: #f99;
  text-align: center;
  padding: 0.5em 0;
}
.diffoscope table.diff tr.ondemand:hover td, .diffoscope div.ondemand-details:hover {
  background: #faa;
  cursor: pointer;
}
"""

SCRIPTS = r"""<script src="%(jquery_url)s"></script>
<script type="text/javascript">
$(function() {
  // activate [+]/[-] controls
  var diffcontrols = $(".diffcontrol");
  $(".diffheader").on('click', function(evt) {
    var control = $(this).find(".diffcontrol");
    var parent = control.parent();
    var target = parent.siblings('table.diff, div.difference, div.comment');
    var orig = target;
    if (evt.shiftKey) {
        var gparent = parent.parent();
        target = gparent.find('table.diff, div.comment, div.difference');
        control = target.parent().not(gparent).find('.diffcontrol');
    }
    if (orig.is(":visible")) {
        target.hide();
        control.text("⊞");
    } else {
        target.show();
        control.text("⊟");
    }
  });
  diffcontrols.attr('title','shift-click to show/hide all children too.');
  diffcontrols.show();
  $(".diffcontrol-nochildren").show();
});
</script>
<style>
.diffoscope .diffheader {
  cursor: pointer;
}
.diffoscope .diffheader:hover .diffcontrol {
  color: #080;
  font-weight: bold;
}
.diffoscope .diffcontrol-double {
  line-height: 250%%;
}
</style>
"""

DIFFNODE_LAZY_LOAD = """<div class="ondemand-details" title="the size refers to the raw diff and includes all children">... <a
href="%(pagename)s.html" target="_blank">open details (total %(size)s)</a> ...</div>
"""

DIFFNODE_LIMIT = """<div class="error">Max HTML report size reached</div>
"""

UD_TABLE_HEADER = """<table class="diff">
<colgroup><col class="colines"/><col class="coldiff"/>
<col class="colines"/><col class="coldiff"/></colgroup>
<tr style="display:none;"><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td></tr>
"""

UD_TABLE_FOOTER = """<tr class="ondemand"><td colspan="4">
... <a href="%(filename)s" target="_blank">%(text)s</a> ...
</td></tr>
"""

UD_TABLE_LIMIT_FOOTER = """<tr class="error"><td colspan="4">
Max %(limit_type)s reached; %(bytes_left)s/%(bytes_total)s bytes (%(percent).2f%%) of diff not shown.
</td></tr>"""

EXPANDED_UD_HEADER = """<div class="difference">"""

EXPANDED_UD_FOOTER = """</div>"""



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017 Maria Glukhova <siammezzze@gmail.com>
# Copyright © 2017, 2019-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import sys
import pytest
import subprocess

from diffoscope.comparators.apk import ApkFile

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least
from ..utils.nonexisting import assert_non_existing

apk1 = load_fixture("test1.apk")
apk2 = load_fixture("test2.apk")
apk3 = load_fixture("test3.apk")


def apktool_version():
    return (
        subprocess.check_output(["apktool", "-version"])
        .decode("utf-8")
        .strip()
    )


def test_identification(apk1):
    assert isinstance(apk1, ApkFile)


def test_no_differences(apk1):
    difference = apk1.compare(apk1)
    assert difference is None


@pytest.fixture
def differences(apk1, apk2):
    return apk1.compare(apk2).details


@pytest.fixture
def differences2(apk1, apk3):
    return apk1.compare(apk3).details


@skip_unless_tools_exist("apktool", "zipinfo")
def test_compare_non_existing(monkeypatch, apk1):
    assert_non_existing(monkeypatch, apk1)


@skip_unless_tools_exist("apktool", "zipinfo")
def test_zipinfo(differences):
    assert differences[0].source1 == "zipinfo {}"
    assert differences[0].source2 == "zipinfo {}"
    assert_diff(differences[0], "apk_zipinfo_expected_diff")


@pytest.mark.xfail(strict=False)
@skip_unless_tools_exist("zipinfo")
@skip_unless_tool_is_at_least("apktool", apktool_version, "2.5.0")
@pytest.mark.skipif(
    sys.version_info < (3, 8), reason="requires Python 3.8 or higher"
)
def test_android_manifest(differences):
    assert differences[1].source1 == "AndroidManifest.xml (decoded)"
    assert differences[1].source2 == "AndroidManifest.xml (decoded)"
    assert_diff(differences[1].details[0], "apk_manifest_expected_diff")


@skip_unless_tools_exist("apktool", "zipinfo")
def test_apk_metadata_source(differences):
    assert differences[2].source1 == "APK metadata"
    assert differences[2].source2 == "APK metadata"


@skip_unless_tools_exist("apktool", "zipinfo")
def test_skip_undecoded_android_manifest(differences):
    assert not any(
        difference.source1 == "original/AndroidManifest.xml"
        for difference in differences
    )
    assert not any(
        difference.source2 == "original/AndroidManifest.xml"
        for difference in differences
    )
    undecoded_manifest = "AndroidManifest.xml (original / undecoded)"
    assert not any(
        difference.source1 == undecoded_manifest for difference in differences
    )
    assert not any(
        difference.source2 == undecoded_manifest for difference in differences
    )


@skip_unless_tools_exist("apktool", "zipinfo")
def test_no_android_manifest(differences2):
    undecoded_manifest = "AndroidManifest.xml (original / undecoded)"
    assert differences2[2].source1 == undecoded_manifest
    assert differences2[2].source2 == undecoded_manifest
    assert (
        differences2[2].comment == "No decoded AndroidManifest.xml "
        "found for one of the APK files."
    )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import logging

from diffoscope.tools import tool_required
from diffoscope.tempfiles import get_temporary_directory

from .utils.file import File
from .utils.archive import Archive
from .utils.command import our_check_output

logger = logging.getLogger(__name__)


class AsarContainer(Archive):
    @tool_required("asar")
    def open_archive(self):
        # Just extract the asar as a whole due to the problems mentioned below
        self._members = []
        self._unpacked = get_temporary_directory(suffix="asar")

        logger.debug("Extracting asar to %s", self._unpacked.name)

        our_check_output(
            [
                "asar",
                "extract",
                os.path.abspath(self.source.path),
                self._unpacked.name,
            ],
            cwd=self._unpacked.name,
        )

        for root, _, files in os.walk(self._unpacked.name):
            current_dir = []

            for filename in sorted(files):
                abspath = os.path.join(root, filename)

                relpath = abspath[len(self._unpacked.name) + 1 :]

                current_dir.append(relpath)

            self._members.extend(current_dir)

        return self

    def close_archive(self):
        self._unpacked.cleanup()

    def get_member_names(self):
        # 'asar list' returns all entries including directories, so it's hard to use
        # rely on extraction to temp directory instead
        return self._members

    def extract(self, member_name, dest_dir):
        # while 'asar' lets us extract single files, it does not let us specify where to output them at all.
        # rely on extraction to temp directory instead
        return os.path.join(self._unpacked.name, member_name)


class AsarFile(File):
    DESCRIPTION = "asar archives"
    CONTAINER_CLASSES = [AsarContainer]
    FILE_TYPE_RE = re.compile(r"^\bElectron ASAR archive\b")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2017, 2020, 2024-2025 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import shutil
import pytest

from diffoscope.comparators.lzma import LzmaFile
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing

lzma1 = load_fixture("test1.lzma")
lzma2 = load_fixture("test2.lzma")


def test_identification(lzma1):
    assert isinstance(lzma1, LzmaFile)


def test_no_differences(lzma1):
    difference = lzma1.compare(lzma1)
    assert difference is None


@pytest.fixture
def differences(lzma1, lzma2):
    return lzma1.compare(lzma2).details


@skip_unless_tools_exist("xz")
def test_content_source(differences):
    assert differences[0].source1 == "test1"
    assert differences[0].source2 == "test2"


@skip_unless_tools_exist("xz")
def test_content_source_without_extension(tmpdir, lzma1, lzma2):
    path1 = str(tmpdir.join("test1"))
    path2 = str(tmpdir.join("test2"))
    shutil.copy(lzma1.path, path1)
    shutil.copy(lzma2.path, path2)
    lzma1 = specialize(FilesystemFile(path1))
    lzma2 = specialize(FilesystemFile(path2))
    difference = lzma1.compare(lzma2).details
    assert difference[0].source1 == "test1-content"
    assert difference[0].source2 == "test2-content"


@skip_unless_tools_exist("xz")
def test_content_diff(differences):
    assert_diff(differences[0], "text_ascii_expected_diff")


@skip_unless_tools_exist("xz")
def test_compare_non_existing(monkeypatch, lzma1):
    assert_non_existing(monkeypatch, lzma1)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import pytest

from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.device import Device
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import load_fixture, get_data, normalize_zeros
from ..utils.tools import skip_unless_tools_exist

text_ascii1 = load_fixture("text_ascii1")


@pytest.fixture
def devnull():
    return specialize(FilesystemFile("/dev/null"))


@pytest.fixture
def differences(devnull, text_ascii1):
    return devnull.compare_bytes(text_ascii1)


@pytest.fixture
def differences_reverse(text_ascii1, devnull):
    return text_ascii1.compare_bytes(devnull)


def test_identification(devnull):
    assert isinstance(devnull, Device)


@skip_unless_tools_exist("xxd")
def test_diff(differences):
    if os.uname()[0] == "FreeBSD":
        expected_diff = get_data("device_expected_diff_freebsd")
    else:
        expected_diff = get_data("device_expected_diff")
    assert normalize_zeros(differences.unified_diff) == expected_diff


@skip_unless_tools_exist("xxd")
def test_diff_reverse(differences_reverse):
    if os.uname()[0] == "FreeBSD":
        expected_diff = get_data("device_expected_diff_reverse_freebsd")
    else:
        expected_diff = get_data("device_expected_diff_reverse")
    assert normalize_zeros(differences_reverse.unified_diff) == expected_diff



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016 Reiner Herrmann <reiner@reiner-h.de>
# Copyright © 2016-2018, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import logging

from diffoscope.exc import RequiredToolNotFound
from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .text import TextFile
from .utils.command import Command

logger = logging.getLogger(__name__)


class Pstotext(Command):
    @tool_required("ps2ascii")
    def cmdline(self):
        return ["ps2ascii", self.path]


class PsFile(TextFile):
    DESCRIPTION = "PostScript documents"
    FILE_TYPE_RE = re.compile(r"^PostScript document\b")

    def compare(self, other, *args, **kwargs):
        differences = super().compare(other, *args, **kwargs)
        details = None
        try:
            details = Difference.from_operation(
                Pstotext, self.path, other.path
            )
        except RequiredToolNotFound:  # noqa
            logger.debug("ps2ascii not found")

        if details:
            differences.add_details([details])
        return differences



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2017, 2020, 2022-2023 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.haskell import HiFile

from ..utils.data import get_data, load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist

haskell1 = load_fixture("test1.hi")
haskell2 = load_fixture("test2.hi")


@skip_unless_tools_exist("ghc")
def test_identification(haskell1):
    if not isinstance(haskell1, HiFile):
        pytest.skip("mismatch between system ghc and fixture")

    assert isinstance(haskell1, HiFile)


def test_no_differences(haskell1):
    assert haskell1.compare(haskell1) is None


@pytest.fixture
def differences(haskell1, haskell2):
    return haskell1.compare(haskell2).details


@skip_unless_tools_exist("ghc")
def test_diff(haskell1, differences):
    if not isinstance(haskell1, HiFile):
        pytest.skip("mismatch between system ghc and fixture")

    with open("tests/data/haskell_expected_diff", "w") as f:
        f.write(differences[0].unified_diff)

    assert_diff(differences[0], "haskell_expected_diff")
